from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from AccessControl.interfaces import IUser
from AccessControl.requestmethod import postonly
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2
from Products.CMFCore.interfaces import IMember
from Products.CMFCore.MemberDataTool import _marker
from Products.CMFCore.MemberDataTool import MemberAdapter as BaseMemberAdapter
from Products.CMFCore.MemberDataTool import MemberDataTool as BaseTool
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.interfaces.capabilities import IAssignRoleCapability
from Products.PlonePAS.interfaces.capabilities import IDeleteCapability
from Products.PlonePAS.interfaces.capabilities import IGroupCapability
from Products.PlonePAS.interfaces.capabilities import IManageCapabilities
from Products.PlonePAS.interfaces.capabilities import IPasswordSetCapability
from Products.PlonePAS.interfaces.group import IGroupManagement
from Products.PlonePAS.interfaces.memberdata import IMemberDataTool
from Products.PlonePAS.interfaces.plugins import IUserManagement
from Products.PlonePAS.interfaces.propertysheets import IMutablePropertySheet
from Products.PluggableAuthService.events import PropertiesUpdated
from Products.PluggableAuthService.interfaces.authservice import IPluggableAuthService
from Products.PluggableAuthService.interfaces.plugins import IPropertiesPlugin
from Products.PluggableAuthService.interfaces.plugins import IRoleAssignerPlugin
from zope.component import adapter
from zope.event import notify
from zope.interface import implementer


@implementer(IMemberDataTool)
class MemberDataTool(BaseTool):
    """PAS-specific implementation of memberdata tool."""

    meta_type = "PlonePAS MemberData Tool"
    security = ClassSecurityInfo()
    toolicon = "tool.gif"

    def __init__(self):
        BaseTool.__init__(self)
        self.portraits = BTreeFolder2(id="portraits")

    def _getPortrait(self, member_id):
        "return member_id's portrait if you can"
        return self.portraits.get(member_id, None)

    def _setPortrait(self, portrait, member_id):
        "store portrait which must be a raw image in _portrais"
        if member_id in self.portraits:
            self.portraits._delObject(member_id)
        self.portraits._setObject(id=member_id, object=portrait)

    def _deletePortrait(self, member_id):
        "remove member_id's portrait"
        if member_id in self.portraits:
            self.portraits._delObject(member_id)

    @security.private
    def pruneMemberDataContents(self):
        """
        Compare the user IDs stored in the member data
        tool with the list in the actual underlying acl_users
        and delete anything not in acl_users
        """
        BaseTool.pruneMemberDataContents(self)
        membertool = getToolByName(self, "portal_membership")
        portraits = self.portraits
        user_list = membertool.listMemberIds()

        for tuple in portraits.items():
            member_id = tuple[0]
            if member_id not in user_list:
                self.portraits._delObject(member_id)

    @security.protected(ManagePortal)
    def purgeMemberDataContents(self):
        """
        Delete ALL MemberData information. This is required for us as we change
        the MemberData class.
        """
        members = self._members

        for tuple in members.items():
            member_name = tuple[0]
            del members[member_name]

        return "Done."

    @security.private
    def updateMemberDataContents(
        self,
    ):
        """Update former MemberData objects to new MemberData objects"""
        count = 0
        members = self._members
        properties = self.propertyIds()

        # Scan members for old MemberData
        for member_name, member_obj in members.items():
            values = {}
            if getattr(member_obj, "_is_new_kind", None):
                continue  # Do not have to upgrade that object

            # Have to upgrade. Create the values mapping.
            for pty_name in properties:
                user_value = getattr(member_obj, pty_name, _marker)
                if user_value is not _marker:
                    values[pty_name] = user_value

            # Wrap a new user object of the RIGHT class
            u = self.acl_users.getUserById(member_name, None)
            if not u:
                continue  # User is not in main acl_users anymore
            self.wrapUser(u)

            # Set its properties
            mbr = self._members.get(member_name, None)
            if not mbr:
                raise RuntimeError(f"Error while upgrading user '{member_name}'.")
            mbr.setProperties(values, force_local=1)
            count += 1

        return count

    @security.private
    def searchMemberDataContents(self, search_param, search_term):
        """
        Search members.
        This is the same as CMFCore except that it doesn't check term case.
        """
        res = []

        search_term = search_term.strip().lower()

        if search_param == "username":
            search_param = "id"

        mtool = getToolByName(self, "portal_membership")

        for member_id in self._members.keys():
            user_wrapper = mtool.getMemberById(member_id)

            if user_wrapper is not None:
                memberProperty = user_wrapper.getProperty
                searched = memberProperty(search_param, None)

                if searched is not None:
                    if searched.strip().lower().find(search_term) != -1:

                        res.append(
                            {
                                "username": memberProperty("id"),
                                "email": memberProperty("email", ""),
                            }
                        )
        return res

    @security.public
    def searchFulltextForMembers(self, s):
        # Note: do NOT add a docstring to this method: it is okay if the method
        # is public, but it must not be published through the web, which is
        # what a docstring would do.
        #
        # Search for members which do have string 's' in name, email or full
        # name (if defined).  This is mainly used for the localrole form.
        s = s.strip().lower()
        mu = getToolByName(self, "portal_membership")

        res = []
        for member in mu.listMembers():
            u = member.getUser()
            if (
                u.getUserName().lower().find(s) != -1
                or member.getProperty("fullname").lower().find(s) != -1
                or member.getProperty("email").lower().find(s) != -1
            ):
                res.append(member)
        return res

    # check to see if we can add users. Need to be careful here
    # so we do not write on read
    def canAddMemberData(self):
        try:
            if self.REQUEST.REQUEST_METHOD != "POST":
                return False
            if (
                getattr(self, "_p_jar", None)
                and len(self._p_jar._registered_objects) > 0
            ):
                # XXX do not write on read
                return True
        except AttributeError:
            pass
        return False

    @postonly
    def deleteMemberData(self, member_id, REQUEST=None):
        """Delete member data of specified member."""
        if IPluggableAuthService.providedBy(self.acl_users):
            # It's a PAS! Whee!
            # XXX: can we safely assume that user name == member_id
            plugins = self._getPlugins()
            prop_managers = plugins.listPlugins(IPropertiesPlugin)
            for mid, prop_manager in prop_managers:
                # Not all PropertiesPlugins support user deletion
                try:
                    prop_manager.deleteUser(member_id)
                except AttributeError:
                    pass

        # we won't always have PlonePAS users, due to acquisition,
        # nor are guaranteed property sheets
        members = self._members
        if member_id in members:
            del members[member_id]
            return 1
        else:
            return 0

    # plugin getter
    def _getPlugins(self):
        return self.acl_users.plugins


InitializeClass(MemberDataTool)


@implementer(IManageCapabilities, IMember)
@adapter(IUser, IMemberDataTool)
class MemberData(BaseMemberAdapter):

    security = ClassSecurityInfo()

    def __init__(self, user, tool):
        BaseMemberAdapter.__init__(self, user, tool)
        self.id = user.getId()

    # setProperties uses setMemberProperties. no need to override.

    def setMemberProperties(self, mapping, force_local=0, force_empty=False):
        """PAS-specific method to set the properties of a
        member. Ignores 'force_local', which is not reliably present.
        """
        sheets = None

        # We could pay attention to force_local here...
        if not IPluggableAuthService.providedBy(self._tool.acl_users):
            # Defer to base impl in absence of PAS, a PAS user, or
            # property sheets
            return BaseMemberAdapter.setMemberProperties(self, mapping)
        else:
            # It's a PAS! Whee!
            user = self.getUser()
            sheets = getattr(user, "getOrderedPropertySheets", lambda: None)()

            # We won't always have PlonePAS users, due to acquisition,
            # nor are guaranteed property sheets
            if not sheets:
                # Defer to base impl if we have a PAS but no property
                # sheets.
                return BaseMemberAdapter.setMemberProperties(self, mapping)

        # If we got this far, we have a PAS and some property sheets.
        # XXX track values set to defer to default impl
        # property routing?
        modified = False
        for k, v in mapping.items():
            if v is None and not force_empty:
                continue
            for sheet in sheets:
                if not sheet.hasProperty(k):
                    continue
                if IMutablePropertySheet.providedBy(sheet):
                    sheet.setProperty(user, k, v)
                    modified = True
                else:
                    break
        if modified:
            self.notifyModified()

        # Trigger PropertiesUpdated event when member properties are updated,
        # excluding user login events
        if not set(mapping.keys()) & {"login_time", "last_login_time"}:
            notify(PropertiesUpdated(self, mapping))

    def getProperty(self, id, default=_marker):
        """PAS-specific method to fetch a user's properties. Looks
        through the ordered property sheets.
        """
        sheets = None
        if not IPluggableAuthService.providedBy(self._tool.acl_users):
            return BaseMemberAdapter.getProperty(self, id)
        else:
            # It's a PAS! Whee!
            user = self.getUser()
            sheets = getattr(user, "getOrderedPropertySheets", lambda: None)()

            # we won't always have PlonePAS users, due to acquisition,
            # nor are guaranteed property sheets
            if not sheets:
                try:
                    return BaseMemberAdapter.getProperty(self, id, default)
                except ValueError:
                    # Zope users don't have PropertySheets,
                    # return an empty string for them if the property
                    # doesn't exists.
                    return ""

        # If we made this far, we found a PAS and some property sheets.
        for sheet in sheets:
            if sheet.hasProperty(id):
                # Return the first one that has the property.
                return sheet.getProperty(id)

        # Couldn't find the property in the property sheets. Try to
        # delegate back to the base implementation.
        return BaseMemberAdapter.getProperty(self, id, default)

    @security.public
    def hasProperty(self, propname):
        """Does the member have the given property?"""
        # Unfortunately, checking for a marker value does not work well:
        # You get a ValueError when the property does not exist.
        try:
            return self.getProperty(propname, _marker) != _marker
        except ValueError:
            return 0

    def getPassword(self):
        """Returns None. Present to avoid NotImplementedError."""
        return None

    # IManageCapabilities methods

    def canDelete(self):
        """True iff user can be removed from the Plone UI."""
        # IUserManagement provides doDeleteUser
        plugins = self._getPlugins()
        managers = plugins.listPlugins(IUserManagement)
        for mid, manager in managers:
            if IDeleteCapability.providedBy(manager) and manager.allowDeletePrincipal(
                self.getId()
            ):
                return True
        return False

    def canPasswordSet(self):
        """True iff user can change password."""
        # IUserManagement provides doChangeUser
        plugins = self._getPlugins()
        managers = plugins.listPlugins(IUserManagement)
        for mid, manager in managers:
            if IPasswordSetCapability.providedBy(manager) and manager.allowPasswordSet(
                self.getId()
            ):
                return True
        return False

    def passwordInClear(self):
        """True iff password can be retrieved in the clear (not hashed.)

        False for PAS. It provides no API for getting passwords,
        though it would be possible to add one in the future.
        """
        return 0

    def _memberdataHasProperty(self, prop_name):
        mdata = getToolByName(self, "portal_memberdata", None)
        if mdata:
            return mdata.hasProperty(prop_name)
        return 0

    def canWriteProperty(self, prop_name):
        """True iff the member/group property named in 'prop_name'
        can be changed.
        """
        if not IPluggableAuthService.providedBy(self._tool.acl_users):
            # not PAS; Memberdata is writable
            return self._memberdataHasProperty(prop_name)
        else:
            # it's PAS
            user = self.getUser()
            sheets = getattr(user, "getOrderedPropertySheets", lambda: None)()
            if not sheets:
                return self._memberdataHasProperty(prop_name)

            for sheet in sheets:
                if not sheet.hasProperty(prop_name):
                    continue
                if IMutablePropertySheet.providedBy(sheet):
                    # BBB for plugins implementing an older version of
                    # IMutablePropertySheet
                    if hasattr(sheet, "canWriteProperty"):
                        return sheet.canWriteProperty(user, prop_name)
                    return True
                else:
                    break  # shadowed by read-only
        return False

    def canAddToGroup(self, group_id):
        """True iff member can be added to group."""
        # IGroupManagement provides IGroupCapability
        plugins = self._getPlugins()
        managers = plugins.listPlugins(IGroupManagement)
        for mid, manager in managers:
            if IGroupCapability.providedBy(manager) and manager.allowGroupAdd(
                self.getId(), group_id
            ):
                return True
        return False

    def canRemoveFromGroup(self, group_id):
        """True iff member can be removed from group."""
        # IGroupManagement provides IGroupCapability
        plugins = self._getPlugins()
        managers = plugins.listPlugins(IGroupManagement)
        for mid, manager in managers:
            if IGroupCapability.providedBy(manager) and manager.allowGroupRemove(
                self.getId(), group_id
            ):
                return True
        return False

    def canAssignRole(self, role_id):
        """True iff member can be assigned role. Role id is string."""
        # IRoleAssignerPlugin provides IAssignRoleCapability
        plugins = self._getPlugins()
        managers = plugins.listPlugins(IRoleAssignerPlugin)
        for mid, manager in managers:
            if IAssignRoleCapability.providedBy(manager) and manager.allowRoleAssign(
                self.getId(), role_id
            ):
                return True
        return False

    @security.private
    def setSecurityProfile(self, password=None, roles=None, domains=None):
        """Set the user's basic security profile"""
        u = self.getUser()

        # The Zope User API is stupid, it should check for None.
        if roles is None:
            roles = list(u.getRoles())
            if "Authenticated" in roles:
                roles.remove("Authenticated")
        if domains is None:
            domains = u.getDomains()

        u.userFolderEditUser(u.getUserId(), password, roles, domains)

    @security.private
    def _getPlugins(self):
        return self._tool._getPlugins()

    @security.public
    def has_permission(self, permission, object):
        return self._user.has_permission(permission, object)

    @security.public
    def getGroups(self):
        return self._user.getGroups()

    @security.public
    def getUserId(self):
        return self.id


InitializeClass(MemberData)



"""
ZODB based user manager with introspection and management interfaces.
"""
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from AccessControl.Permissions import manage_users as ManageUsers
from App.special_dtml import DTMLFile
from AuthEncoding import AuthEncoding
from Products.PlonePAS.interfaces.capabilities import IDeleteCapability
from Products.PlonePAS.interfaces.capabilities import IPasswordSetCapability
from Products.PlonePAS.interfaces.plugins import IUserIntrospection
from Products.PlonePAS.interfaces.plugins import IUserManagement
from Products.PluggableAuthService.events import CredentialsUpdated
from Products.PluggableAuthService.plugins.ZODBUserManager import (
    ZODBUserManager as BasePlugin,
)
from Products.PluggableAuthService.utils import createViewName
from zope.event import notify
from zope.interface import implementer


manage_addUserManagerForm = DTMLFile("../zmi/UserManagerForm", globals())


def manage_addUserManager(dispatcher, id, title=None, REQUEST=None):
    """Add a UserManager to a Pluggable Auth Service."""

    pum = UserManager(id, title)
    dispatcher._setObject(pum.getId(), pum)

    if REQUEST is not None:
        REQUEST["RESPONSE"].redirect(
            "%s/manage_workspace"
            "?manage_tabs_message="
            "UserManager+added." % dispatcher.absolute_url()
        )


@implementer(
    IUserManagement, IUserIntrospection, IDeleteCapability, IPasswordSetCapability
)
class UserManager(BasePlugin):
    """PAS plugin for managing users. (adds write API)"""

    meta_type = "User Manager"
    security = ClassSecurityInfo()

    @security.protected(ManageUsers)
    def addUser(self, user_id, login_name, password):
        """Original ZODBUserManager.addUser, modified to check if
        incoming password is already encypted.

        This support clean migration from default user source.
        Should go into PAS.
        """
        if self._user_passwords.get(user_id) is not None:
            raise KeyError("Duplicate user ID: %s" % user_id)

        if self._login_to_userid.get(login_name) is not None:
            raise KeyError("Duplicate login name: %s" % login_name)

        if not AuthEncoding.is_encrypted(password):
            password = AuthEncoding.pw_encrypt(password)
        self._user_passwords[user_id] = password
        self._login_to_userid[login_name] = user_id
        self._userid_to_login[user_id] = login_name

        # enumerateUsers return value has changed
        view_name = createViewName("enumerateUsers")
        self.ZCacheable_invalidate(view_name=view_name)

    # User Management interface

    @security.private
    def doDeleteUser(self, userid):
        """Given a user id, delete that user"""
        return self.removeUser(userid)

    @security.private
    def doChangeUser(self, principal_id, password):
        """Change a user's password"""
        if self._user_passwords.get(principal_id) is None:
            raise RuntimeError("User does not exist: %s" % principal_id)
        self._user_passwords[principal_id] = AuthEncoding.pw_encrypt(password)
        notify(CredentialsUpdated(self.getUserById(principal_id), password))

    # implement interfaces IDeleteCapability, IPasswordSetCapability

    @security.public
    def allowDeletePrincipal(self, principal_id):
        """True iff this plugin can delete a certain user/group.
        This is true if this plugin manages the user.
        """
        if self._user_passwords.get(principal_id) is not None:
            return 1
        return 0

    @security.public
    def allowPasswordSet(self, principal_id):
        """True iff this plugin can set the password a certain user.
        This is true if this plugin manages the user.
        """
        return self.allowDeletePrincipal(principal_id)

    # User Introspection interface

    @security.protected(ManageUsers)
    def getUserIds(self):
        """
        Return a list of user ids
        """
        return self.listUserIds()

    @security.protected(ManageUsers)
    def getUserNames(self):
        """
        Return a list of usernames
        """
        return [x["login_name"] for x in self.listUserInfo()]

    @security.protected(ManageUsers)
    def getUsers(self):
        """
        Return a list of users
        """
        uf = self.acl_users
        return [uf.getUserById(x) for x in self.getUserIds()]


InitializeClass(UserManager)



from plone.testing import layered
from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_FUNCTIONAL_TESTING

import doctest
import re
import unittest


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(
        layered(
            doctest.DocFileSuite(
                "cookie_auth.rst",
                package="Products.PlonePAS.tests",
                optionflags=doctest.ELLIPSIS,
            ),
            layer=PRODUCTS_PLONEPAS_FUNCTIONAL_TESTING,
        )
    )
    suite.addTest(
        doctest.DocTestSuite(
            "Products.PlonePAS.utils",
            optionflags=doctest.ELLIPSIS,
        )
    )
    return suite



from AccessControl import Permissions
from AccessControl import Unauthorized
from Acquisition import aq_base
from Acquisition import aq_parent
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.plugins.group import PloneGroup
from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING
from Products.PlonePAS.tools.groupdata import GroupData
from Products.PluggableAuthService.interfaces.events import IGroupDeletedEvent
from zope.component import adapter
from zope.component import getGlobalSiteManager

import unittest


def sortTuple(t):
    l = list(t)
    l.sort()
    return tuple(l)


class TestGroupsTool(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.gt = getToolByName(self.portal, "portal_groups")
        self.gd = getToolByName(self.portal, "portal_groupdata")

        self.group_id = "group1"
        # Create a new Group
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.gt.addGroup(
            self.group_id,
            ["Reviewer"],
            [],
            {"email": "group1@host.com", "title": "Group #1"},
        )

    def test_get_group(self):
        # Use PAS (monkeypatched) API method to get a group by id.
        group = self.portal.acl_users.getGroup(self.group_id)
        self.assertFalse(group is None)

        # Should be wrapped into the GroupManagement, which is wrapped
        # into the PAS.
        got = aq_base(aq_parent(aq_parent(group)))
        expected = aq_base(self.portal.acl_users)
        self.assertEqual(got, expected)

        self.assertTrue(isinstance(group, PloneGroup))

    def test_get_group_by_id(self):
        # Use tool way of getting group by id. This returns a
        # GroupData object wrapped by the group
        group = self.gt.getGroupById(self.group_id)
        self.assertFalse(group is None)
        self.assertTrue(isinstance(group, GroupData))
        self.assertTrue(isinstance(aq_parent(group), PloneGroup))

    def test_edit_group(self):
        # Use the tool way to edit a group.
        properties = {"email": "group1@host2.com", "title": "Group #1 new title"}
        self.gt.editGroup(self.group_id, roles=["Manager"], **properties)

        # test edition of roles and properties
        group = self.gt.getGroupById(self.group_id)
        self.assertTrue(group.has_role("Manager"))
        self.assertEqual(group.getProperty("email"), properties["email"])
        self.assertEqual(group.getProperty("title"), properties["title"])

        # test for empty list of roles
        self.gt.editGroup(self.group_id, roles=[])
        self.assertTrue(group.has_role("Authenticated"))

        # test edition of group groups
        self.gt.editGroup(self.group_id, groups=["Reviewers"], **properties)
        group = self.gt.getGroupById(self.group_id)
        self.assertTrue("Reviewers" in group.getGroups())


class TestMethodProtection(unittest.TestCase):
    # GroupData has wrong security declarations

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.groups = self.portal.portal_groups
        self.groups.addGroup("foo")
        self.groupdata = self.groups.getGroupById("foo")

    def testAnonAddMember(self):
        logout()
        self.assertRaises(Unauthorized, self.groupdata.addMember, TEST_USER_ID)

    def testAnonRemoveMember(self):
        logout()
        self.assertRaises(Unauthorized, self.groupdata.removeMember, TEST_USER_ID)

    def testMemberAddMember(self):
        self.assertRaises(Unauthorized, self.groupdata.addMember, TEST_USER_ID)

    def testMemberRemoveMember(self):
        self.assertRaises(Unauthorized, self.groupdata.removeMember, TEST_USER_ID)

    def testManagerAddMember(self):
        self.portal.manage_role("Member", [Permissions.manage_users])
        self.groupdata.addMember(TEST_USER_ID)

    def testManagerRemoveMember(self):
        self.portal.manage_role("Member", [Permissions.manage_users])
        self.groupdata.addMember(TEST_USER_ID)
        self.groupdata.removeMember(TEST_USER_ID)


class TestGroupsToolIntegration(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.membership = self.portal.portal_membership
        self.acl_users = self.portal.acl_users
        self.groups = self.portal.portal_groups
        self.folder = self.portal["folder"]

        if "auto_group" in self.acl_users:
            self.acl_users.manage_delObjects(["auto_group"])

        # Nuke Administators and Reviewers groups added in 2.1a2 migrations
        # (and any other migrated-in groups) to avoid test confusion
        self.groups.removeGroups(self.groups.listGroupIds())

    def testAddGroup(self):
        self.groups.addGroup("foo", [], [])
        self.assertEqual(self.groups.listGroupIds(), ["foo"])

    def testGetGroupById(self):
        self.groups.addGroup("foo", [], [])
        g = self.groups.getGroupById("foo")
        self.assertNotEqual(g, None)

    def testGetBadGroupById(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(g, None)

    def testGroupByIdIsWrapped(self):
        self.groups.addGroup("foo", [], [])
        g = self.groups.getGroupById("foo")
        self.assertEqual(g.__class__.__name__, "GroupData")
        self.assertEqual(aq_parent(g).__class__.__name__, "PloneGroup")
        self.assertEqual(aq_parent(aq_parent(g)).__class__.__name__, "GroupManager")

    def testEditGroup(self):
        self.groups.addGroup(
            "foo",
        )
        self.groups.editGroup("foo", roles=["Reviewer"])
        g = self.groups.getGroupById("foo")
        self.assertEqual(sortTuple(g.getRoles()), ("Authenticated", "Reviewer"))

    def testEditBadGroup(self):
        # Error type depends on the user folder...
        try:
            self.groups.editGroup("foo", [], [])
        except (KeyError, ValueError):
            pass  # Ok, this is the wanted behaviour
        else:
            self.fail("Should have raised KeyError or ValueError")

    def testRemoveGroups(self):
        self.groups.addGroup("foo", [], [])
        self.groups.removeGroups(["foo"])
        self.assertEqual(len(self.groups.listGroupIds()), 0)

    def testRemoveGroupDelEvent(self):
        eventsFired = []

        @adapter(IGroupDeletedEvent)
        def gotDeletion(event):
            eventsFired.append(event)

        gsm = getGlobalSiteManager()
        gsm.registerHandler(gotDeletion)
        self.groups.addGroup("foo", [], [])
        self.groups.removeGroups(["foo"])
        self.assertEqual(len(eventsFired), 1)
        self.assertEqual(eventsFired[0].principal, "foo")
        gsm.unregisterHandler(gotDeletion)

    def testListGroupIds(self):
        self.groups.addGroup("foo", [], [])
        self.groups.addGroup("bar", [], [])
        grps = self.groups.listGroupIds()
        grps.sort()
        self.assertEqual(grps, ["bar", "foo"])

    def testGetGroupsByUserId(self):
        self.groups.addGroup("foo", [], [])
        self.acl_users.userSetGroups(TEST_USER_ID, groupnames=["foo"])
        gs = self.groups.getGroupsByUserId(TEST_USER_ID)
        self.assertEqual(gs[0].getId(), "foo")

    def testGroupsByUserIdAreWrapped(self):
        self.groups.addGroup("foo", [], [])
        self.acl_users.userSetGroups(TEST_USER_ID, groupnames=["foo"])
        gs = self.groups.getGroupsByUserId(TEST_USER_ID)
        self.assertEqual(gs[0].__class__.__name__, "GroupData")
        self.assertEqual(aq_parent(gs[0]).__class__.__name__, "PloneGroup")
        self.assertEqual(aq_parent(aq_parent(gs[0])).__class__.__name__, "GroupManager")

    def testListGroups(self):
        self.groups.addGroup("foo", [], [])
        self.groups.addGroup("bar", [], [])
        gs = self.groups.listGroups()
        self.assertEqual(gs[0].getId(), "bar")
        self.assertEqual(gs[1].getId(), "foo")

    def testListedGroupsAreWrapped(self):
        self.groups.addGroup("foo", [], [])
        gs = self.groups.listGroups()
        self.assertEqual(gs[0].__class__.__name__, "GroupData")
        self.assertEqual(aq_parent(gs[0]).__class__.__name__, "PloneGroup")
        self.assertEqual(aq_parent(aq_parent(gs[0])).__class__.__name__, "GroupManager")

    def testSetGroupOwnership(self):
        self.groups.addGroup("foo", [], [])
        self.folder.invokeFactory("Document", "doc")
        doc = self.folder.doc
        g = self.groups.getGroupById("foo")
        self.groups.setGroupOwnership(g, doc)
        self.assertEqual(doc.getOwnerTuple()[1], "foo")
        self.assertEqual(doc.get_local_roles_for_userid("foo"), ("Owner",))
        self.assertEqual(doc.get_local_roles_for_userid(TEST_USER_ID), ("Owner",))

    def testWrapGroup(self):
        self.groups.addGroup("foo", [], [])
        g = self.acl_users.getGroup("foo")
        self.assertEqual(g.__class__.__name__, "PloneGroup")
        g = self.groups.wrapGroup(g)
        self.assertEqual(g.__class__.__name__, "GroupData")
        self.assertEqual(aq_parent(g).__class__.__name__, "PloneGroup")
        self.assertEqual(aq_parent(aq_parent(g)).__class__.__name__, "GroupManager")

    def testGetGroupInfo(self):
        self.groups.addGroup("foo", title="Foo", description="Bar", email="foo@foo.com")
        info = self.groups.getGroupInfo("foo")
        self.assertEqual(info.get("title"), "Foo")
        self.assertEqual(info.get("description"), "Bar")
        self.assertEqual(info.get("email"), None)  # No email!

    def testGetGroupInfoAsAnonymous(self):
        self.groups.addGroup("foo", title="Foo", description="Bar")
        logout()
        info = self.groups.restrictedTraverse("getGroupInfo")("foo")
        self.assertEqual(info.get("title"), "Foo")
        self.assertEqual(info.get("description"), "Bar")

    def testGetBadGroupInfo(self):
        info = self.groups.getGroupInfo("foo")
        self.assertEqual(info, None)



from PIL import Image


PROJECTNAME = "PlonePAS"
GLOBALS = globals()

DEFAULT_CHALLENGE_PROTOCOL = ["http"]
DEFAULT_PROTO_MAPPING = {
    "WebDAV": DEFAULT_CHALLENGE_PROTOCOL,
    "FTP": DEFAULT_CHALLENGE_PROTOCOL,
    "XML-RPC": DEFAULT_CHALLENGE_PROTOCOL,
}

# Settings for member image resize quality
HAS_PIL = True  # remove in Version 8
PIL_SCALING_ALGO = Image.Resampling.LANCZOS
PIL_QUALITY = 88
MEMBER_IMAGE_SCALE = (75, 100)
IMAGE_SCALE_PARAMS = {
    "scale": MEMBER_IMAGE_SCALE,
    "quality": PIL_QUALITY,
    "algorithm": PIL_SCALING_ALGO,
    "default_format": "PNG",
}



from Products.PlonePAS.interfaces.events import IUserInitialLoginInEvent
from Products.PluggableAuthService.events import PASEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedInEvent
from Products.PluggableAuthService.interfaces.events import IUserLoggedOutEvent
from zope.interface import implementer


@implementer(IUserLoggedInEvent)
class UserLoggedInEvent(PASEvent):
    """Plone Implementation of the logged in event

    PAS Event
    """


@implementer(IUserInitialLoginInEvent)
class UserInitialLoginInEvent(UserLoggedInEvent):
    """Implementation of the initial logged in event

    Plone only event!
    """


@implementer(IUserLoggedOutEvent)
class UserLoggedOutEvent(PASEvent):
    """Plone Implementation of the logged out event

    PAS Event
    """



"""
ZODB Group Implementation with basic introspection and
management (ie. rw) capabilities.

"""
from .ufactory import PloneUser
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from App.special_dtml import DTMLFile
from BTrees.OOBTree import OOBTree
from BTrees.OOBTree import OOSet
from Products.PlonePAS.interfaces.capabilities import IDeleteCapability
from Products.PlonePAS.interfaces.capabilities import IGroupCapability
from Products.PlonePAS.interfaces.group import IGroupIntrospection
from Products.PlonePAS.interfaces.group import IGroupManagement
from Products.PluggableAuthService.interfaces.plugins import IGroupEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins import IPropertiesPlugin
from Products.PluggableAuthService.interfaces.plugins import IRolesPlugin
from Products.PluggableAuthService.PluggableAuthService import (
    _SWALLOWABLE_PLUGIN_EXCEPTIONS,
)
from Products.PluggableAuthService.plugins.ZODBGroupManager import ZODBGroupManager
from zope.interface import implementer

import logging


manage_addGroupManagerForm = DTMLFile("../zmi/GroupManagerForm", globals())
logger = logging.getLogger("PlonePAS")


def manage_addGroupManager(self, id, title="", RESPONSE=None):
    """
    Add a zodb group manager with management and introspection
    capabilities to pas.
    """
    grum = GroupManager(id, title)

    self._setObject(grum.getId(), grum)

    if RESPONSE is not None:
        return RESPONSE.redirect("manage_workspace")


@implementer(IGroupManagement, IGroupIntrospection, IGroupCapability, IDeleteCapability)
class GroupManager(ZODBGroupManager):

    meta_type = "Group Manager"
    security = ClassSecurityInfo()

    def __init__(self, *args, **kw):
        ZODBGroupManager.__init__(self, *args, **kw)
        # reverse index of groups->principal
        self._group_principal_map = OOBTree()

    #################################
    # overrides to ease group principal lookups for introspection api

    def addGroup(self, group_id, *args, **kw):
        ZODBGroupManager.addGroup(self, group_id, *args, **kw)
        self._group_principal_map[group_id] = OOSet()
        return True

    def removeGroup(self, group_id):
        ZODBGroupManager.removeGroup(self, group_id)
        del self._group_principal_map[group_id]
        return True

    def addPrincipalToGroup(self, principal_id, group_id):
        ZODBGroupManager.addPrincipalToGroup(self, principal_id, group_id)
        self._group_principal_map[group_id].insert(principal_id)
        return True

    def removePrincipalFromGroup(self, principal_id, group_id):
        already = ZODBGroupManager.removePrincipalFromGroup(
            self, principal_id, group_id
        )
        if already:
            self._group_principal_map[group_id].remove(principal_id)
        return True

    #################################
    # overrides for api matching/massage

    def updateGroup(self, group_id, title=None, description=None):
        ZODBGroupManager.updateGroup(
            self, group_id, title=title, description=description
        )
        return True

    #################################
    # introspection interface

    def getGroupById(self, group_id, default=None):
        plugins = self._getPAS()._getOb("plugins")
        title = None
        if group_id not in self.getGroupIds():
            return default

        return self._findGroup(plugins, group_id, title)

    def getGroups(self):
        return [self.getGroupById(group_id) for group_id in self.getGroupIds()]

    def getGroupIds(self):
        return self.listGroupIds()

    def getGroupMembers(self, group_id):
        return tuple(self._group_principal_map.get(group_id, ()))

    #################################
    # capabilties interface impls.

    @security.public
    def allowDeletePrincipal(self, principal_id):
        """True iff this plugin can delete a certain group.
        This is true if this plugin manages the group.
        """
        if self._groups.get(principal_id) is not None:
            return 1
        return 0

    def getGroupInfo(self, group_id):
        """Over-ride parent to not explode when getting group info dict by
        group id."""
        return self._groups.get(group_id, None)

    def allowGroupAdd(self, user_id, group_id):
        """True iff this plugin will allow adding a certain user to a
        certain group."""
        present = self.getGroupInfo(group_id)
        # if we have a group, we can add users to it
        # slightly naive, but should be okay.
        if present:
            return 1

        return 0

    def allowGroupRemove(self, user_id, group_id):
        """True iff this plugin will allow removing a certain user from a
        certain group."""
        present = self.getGroupInfo(group_id)
        # if we don't have a group, we can't do anything
        if not present:
            return 0

        group_members = self.getGroupMembers(group_id)
        if user_id in group_members:
            return 1
        return 0

    #################################
    # group wrapping mechanics

    @security.private
    def _createGroup(self, plugins, group_id, name):
        """Create group object. For users, this can be done with a
        plugin, but I don't care to define one for that now. Just uses
        PloneGroup.  But, the code's still here, just commented out.
        This method based on PluggableAuthervice._createUser
        """
        return PloneGroup(group_id, name).__of__(self)

    @security.private
    def _findGroup(self, plugins, group_id, title=None, request=None):
        """group_id -> decorated_group
        This method based on PluggableAuthService._findGroup
        """
        group = self._createGroup(plugins, group_id, title)

        propfinders = plugins.listPlugins(IPropertiesPlugin)
        for propfinder_id, propfinder in propfinders:

            data = propfinder.getPropertiesForUser(group, request)
            if data:
                group.addPropertysheet(propfinder_id, data)

        groups = self._getPAS()._getGroupsForPrincipal(group, request, plugins=plugins)
        group._addGroups(groups)

        rolemakers = plugins.listPlugins(IRolesPlugin)

        for rolemaker_id, rolemaker in rolemakers:
            roles = rolemaker.getRolesForPrincipal(group, request)
            if roles:
                group._addRoles(roles)

        group._addRoles(["Authenticated"])

        return group.__of__(self)

    @security.private
    def _verifyGroup(self, plugins, group_id=None, title=None):

        """group_id -> boolean
        This method based on PluggableAuthService._verifyUser
        """
        criteria = {}

        if group_id is not None:
            criteria["id"] = group_id
            criteria["exact_match"] = True

        if title is not None:
            criteria["title"] = title

        if criteria:
            enumerators = plugins.listPlugins(IGroupEnumerationPlugin)

            for enumerator_id, enumerator in enumerators:
                try:
                    info = enumerator.enumerateGroups(**criteria)

                    if info:
                        return info[0]["id"]

                except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                    logger.info(
                        "PluggableAuthService: GroupEnumerationPlugin %s " "error",
                        enumerator_id,
                        exc_info=1,
                    )

        return 0


InitializeClass(GroupManager)


class PloneGroup(PloneUser):
    """Plone expects a user to come, with approximately the same
    behavior as a user.
    """

    security = ClassSecurityInfo()
    _isGroup = True

    def getId(self, unprefixed=None):
        """-> user ID
        Modified to accept silly GRUF param.
        """
        return self._id

    @security.private
    def getMemberIds(self, transitive=1):
        """Return member ids of this group, including or not
        transitive groups.
        """
        # acquired from the groups_source
        plugins = self._getPAS().plugins
        introspectors = plugins.listPlugins(IGroupIntrospection)
        members = []
        for iid, introspector in introspectors:
            try:
                members.extend(list(introspector.getGroupMembers(self.getId())))
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                logger.info(
                    "PluggableAuthService: getGroupMembers %s error", iid, exc_info=1
                )

        return members

    @security.public
    def addMember(self, id):
        """Add the existing member with the given id to the group"""
        self.addPrincipalToGroup(id, self.getId())

    @security.public
    def removeMember(self, id):
        """Remove the member with the provided id from the group."""
        self.removePrincipalFromGroup(id, self.getId())

    @security.public
    def getRolesInContext(self, object):
        """Since groups can't actually log in, do nothing."""
        return []

    @security.public
    def allowed(self, object, object_roles=None):
        """Since groups can't actually log in, do nothing."""
        return 0


InitializeClass(PloneGroup)



"""
Upgrade steps specific to Plone's use of PAS.
"""

from Products.PlonePAS.plugins import cookie_handler
from Products.PluggableAuthService.plugins import CookieAuthHelper

import logging


logger = logging.getLogger(__name__)


def from4to5_fix_zope_root(context):
    """
    Fix broken Zope root `/acl_users/` plugins.
    """
    root = context.getPhysicalRoot()
    pas = root.acl_users.manage_addProduct["PluggableAuthService"]
    # Identify which interfaces should be considered PAS plugin interfaces
    plugin_ifaces = [
        plugin_type_info["interface"]
        for plugin_type_info in root.acl_users.plugins.listPluginTypeInfo()
    ]
    broken_meta_type = cookie_handler.ExtendedCookieAuthHelper.meta_type
    broken_plugins = root.acl_users.objectValues(broken_meta_type)
    for broken_plugin in broken_plugins:
        # Collect properties from old/broken plugin
        kwargs = dict(
            id=broken_plugin.id,
            title=broken_plugin.title,
            cookie_name=broken_plugin.cookie_name,
        )
        # Which PAS plugin interfaces has this plugin been activated for
        active_ifaces = [
            plugin_iface
            for plugin_iface in plugin_ifaces
            if plugin_iface.providedBy(broken_plugin)
            and broken_plugin.id in root.acl_users.plugins.listPluginIds(plugin_iface)
        ]
        # Delete the old/broken plugin
        logger.info(
            "Deleting broken %r plugin: %r",
            broken_meta_type,
            "/".join(broken_plugin.getPhysicalPath()),
        )
        root.acl_users.manage_delObjects([broken_plugin.id])
        # Add the correct plugin
        logger.info(
            "Adding working %r plugin: %r",
            CookieAuthHelper.CookieAuthHelper.meta_type,
            "/".join(broken_plugin.getPhysicalPath()),
        )
        pas.addCookieAuthHelper(**kwargs)
        # Restore activated plugin interfaces
        for plugin_iface in active_ifaces:
            root.acl_users.plugins.activatePlugin(
                plugin_iface,
                kwargs["id"],
            )



from AccessControl.Permission import addPermission


AddGroups = "Add Groups"
addPermission(AddGroups, default_roles=("Manager",))

ManageGroups = "Manage Groups"
addPermission(ManageGroups, default_roles=("Manager",))

ViewGroups = "View Groups"
addPermission(ViewGroups, default_roles=("Manager", "Owner", "Member"))

DeleteGroups = "Delete Groups"
addPermission(DeleteGroups, default_roles=("Manager",))

SetGroupOwnership = "Set Group Ownership"
addPermission(SetGroupOwnership, default_roles=("Manager", "Owner"))



from zope.deprecation import deprecation

import logging


logger = logging.getLogger("PlonePAS")

PATTERN = "__PlonePAS_%s__"


def call(self, __name__, *args, **kw):
    return getattr(self, PATTERN % __name__)(*args, **kw)


WRAPPER = "__PlonePAS_is_wrapper_method__"
ADDED = "__PlonePAS_is_added_method__"
ORIG_NAME = "__PlonePAS_original_method_name__"

_marker = dict()  # type: ignore


def isWrapperMethod(meth):
    return getattr(meth, WRAPPER, False)


def wrap_method(
    klass, name, method, pattern=PATTERN, add=False, roles=None, deprecated=False
):
    """takes a method and set it to a class. Annotates with hints what happened."""
    new_name = pattern % name
    if not add:
        old_method = getattr(klass, name)
        if isWrapperMethod(old_method):
            logger.warn(
                "PlonePAS: *NOT* wrapping already wrapped method at "
                "{}.{}".format(klass.__name__, name)
            )

            return
        logger.debug("PlonePAS: Wrapping method at %s.%s", klass.__name__, name)
        setattr(klass, new_name, old_method)
        setattr(method, ORIG_NAME, new_name)
        setattr(method, WRAPPER, True)
        setattr(method, ADDED, False)
    else:
        logger.debug("PlonePAS: Adding method at %s.%s", klass.__name__, name)
        setattr(method, WRAPPER, False)
        setattr(method, ADDED, True)

    if deprecated:
        setattr(klass, name, deprecation.deprecated(method, deprecated))
    else:
        setattr(klass, name, method)

    if roles is not None:
        roles_attr = f"{name}__roles__"
        logger.debug(
            "PlonePAS: Setting new permission roles at {}.{}".format(
                klass.__name__, name
            )
        )
        setattr(klass, roles_attr, roles)


def unwrap_method(klass, name):
    # seems to be dead code, nowwhere used nor tested
    old_method = getattr(klass, name)
    if not isWrapperMethod(old_method):
        raise ValueError(
            "Trying to unwrap non-wrapped " "method at %s.%s" % (klass.__name__, name)
        )
    orig_name = getattr(old_method, ORIG_NAME)
    new_method = getattr(klass, orig_name)
    delattr(klass, orig_name)
    setattr(klass, name, new_method)



"""
Mutable Property Provider
"""
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from App.special_dtml import DTMLFile
from BTrees.OOBTree import OOBTree
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.interfaces.plugins import IMutablePropertiesPlugin
from Products.PlonePAS.sheet import MutablePropertySheet
from Products.PlonePAS.sheet import validateValue
from Products.PlonePAS.utils import safe_unicode
from Products.PluggableAuthService.interfaces.plugins import IPropertiesPlugin
from Products.PluggableAuthService.interfaces.plugins import IUserEnumerationPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.UserPropertySheet import _guessSchema
from ZODB.PersistentMapping import PersistentMapping
from zope.i18nmessageid import MessageFactory
from zope.interface import implementer

import copy


_ = MessageFactory("plone")


def manage_addZODBMutablePropertyProvider(
    self, id, title="", RESPONSE=None, schema=None, **kw
):
    """
    Create an instance of a mutable property manager.
    """
    o = ZODBMutablePropertyProvider(id, title, schema, **kw)
    self._setObject(o.getId(), o)

    if RESPONSE is not None:
        RESPONSE.redirect("manage_workspace")


manage_addZODBMutablePropertyProviderForm = DTMLFile(
    "../zmi/MutablePropertyProviderForm", globals()
)


def isStringType(data):
    return isinstance(data, str) or isinstance(data, str)


@implementer(IPropertiesPlugin, IUserEnumerationPlugin, IMutablePropertiesPlugin)
class ZODBMutablePropertyProvider(BasePlugin):
    """Storage for mutable properties in the ZODB for users/groups.

    API sounds like it's only for users, but groups work as well.
    """

    meta_type = "ZODB Mutable Property Provider"

    security = ClassSecurityInfo()

    def __init__(self, id, title="", schema=None, **kw):
        """Create in-ZODB mutable property provider.

        Provide a schema either as a list of (name,type,value) tuples
        in the 'schema' parameter or as a series of keyword parameters
        'name=value'. Types will be guessed in this case.

        The 'value' is meant as the default value, and will be used
        unless the user provides data.

        If no schema is provided by constructor, the properties of the
        portal_memberdata object will be used.

        Types available: string, text, boolean, int, long, float, lines, date
        """
        self.id = id
        self.title = title
        self._storage = OOBTree()

        # calculate schema and default values
        defaultvalues = {}
        if not schema and not kw:
            schema = ()
        elif not schema and kw:
            schema = _guessSchema(kw)
            defaultvalues = kw
        else:
            valuetuples = [(name, value) for name, type, value in schema]
            schema = [(name, type) for name, type, value in schema]
            for name, value in valuetuples:
                defaultvalues[name] = value
        self._schema = tuple(schema)
        self._defaultvalues = defaultvalues

    def _getSchema(self, isgroup=None):
        # this could probably stand to be cached
        datatool = isgroup and "portal_groupdata" or "portal_memberdata"

        schema = self._schema
        if not schema:
            # if no schema is provided, use portal_memberdata properties
            schema = ()
            mdtool = getToolByName(self, datatool, None)
            # Don't fail badly if tool is not available.
            if mdtool is not None:
                mdschema = mdtool.propertyMap()
                schema = [(elt["id"], elt["type"]) for elt in mdschema]
        return schema

    def _getDefaultValues(self, isgroup=None):
        """Returns a dictionary mapping of property names to default values.
        Defaults to portal_*data tool if necessary.
        """
        datatool = isgroup and "portal_groupdata" or "portal_memberdata"

        defaultvalues = self._defaultvalues
        if not self._schema:
            # if no schema is provided, use portal_*data properties
            defaultvalues = {}
            mdtool = getToolByName(self, datatool, None)
            # Don't fail badly if tool is not available.
            if mdtool is not None:
                # we rely on propertyMap and propertyItems mapping
                mdvalues = mdtool.propertyItems()
                for name, value in mdvalues:
                    # For selection types the default value is the name of a
                    # method which returns the possible values. There is no way
                    # to set a default value for those types.
                    ptype = mdtool.getPropertyType(name)
                    if ptype == "selection":
                        defaultvalues[name] = ""
                    elif ptype == "multiple selection":
                        defaultvalues[name] = []
                    else:
                        defaultvalues[name] = value

            # ALERT! if someone gives their *_data tool a title, and want a
            #        title as a property of the user/group (and groups do by
            #        default) we don't want them all to have this title, since
            #        a title is used in the UI if it exists
            if defaultvalues.get("title"):
                defaultvalues["title"] = ""
        return defaultvalues

    @security.private
    def getPropertiesForUser(self, user, request=None):
        """Get property values for a user or group.
        Returns a dictionary of values or a PropertySheet.

        This implementation will always return a MutablePropertySheet.

        NOTE: Must always return something, or else the property sheet
        won't get created and this will screw up portal_memberdata.
        """
        isGroup = getattr(user, "isGroup", lambda: None)()

        data = self._storage.get(user.getId())
        defaults = self._getDefaultValues(isGroup)

        # provide default values where missing
        if not data:
            data = {}
        for key, val in defaults.items():
            if key not in data:
                data[key] = val

        return MutablePropertySheet(self.id, schema=self._getSchema(isGroup), **data)

    @security.private
    def setPropertiesForUser(self, user, propertysheet):
        """Set the properties of a user or group based on the contents of a
        property sheet.
        """
        isGroup = getattr(user, "isGroup", lambda: None)()

        properties = dict(propertysheet.propertyItems())

        for name, property_type in self._getSchema(isGroup) or ():
            if name in properties and not validateValue(
                property_type, properties[name]
            ):
                raise ValueError(
                    "Invalid value: {} does not conform to {}".format(
                        name, property_type
                    )
                )

        allowed_prop_keys = [pn for pn, pt in self._getSchema(isGroup) or ()]
        if allowed_prop_keys:
            prop_names = set(properties.keys()) - set(allowed_prop_keys)
            if prop_names:
                raise ValueError("Unknown Properties: %r" % prop_names)

        userid = user.getId()
        userprops = self._storage.get(userid)
        properties.update({"isGroup": isGroup})
        if userprops is not None:
            userprops.update(properties)
            # notify persistence machinery of change
            self._storage[userid] = self._storage[userid]
        else:
            self._storage.insert(user.getId(), properties)

    @security.private
    def deleteUser(self, user_id):
        """Delete all user properties"""
        # Do nothing if an unknown user_id is given
        try:
            del self._storage[user_id]
        except KeyError:
            pass

    @security.private
    def testMemberData(self, memberdata, criteria, exact_match=False):
        """Test if a memberdata matches the search criteria."""
        for (key, value) in criteria.items():
            testvalue = memberdata.get(key, None)
            if testvalue is None:
                return False

            if isStringType(testvalue):
                testvalue = safe_unicode(testvalue.lower())
            if isStringType(value):
                value = safe_unicode(value.lower())

            if exact_match:
                if value != testvalue:
                    return False
            else:
                try:
                    if value not in testvalue:
                        return False
                except TypeError:
                    # Fall back to exact match if we can check for
                    # sub-component
                    if value != testvalue:
                        return False

        return True

    @security.private
    def enumerateUsers(self, id=None, login=None, exact_match=False, **kw):
        """See IUserEnumerationPlugin."""
        plugin_id = self.getId()

        # This plugin can't search for a user by id or login, because there is
        # no such keys in the storage (data dict in the comprehensive list)
        # If kw is empty or not, we continue the search.
        if id is not None or login is not None:
            return ()

        criteria = copy.copy(kw)

        users = [
            (user, data)
            for (user, data) in self._storage.items()
            if self.testMemberData(data, criteria, exact_match)
            and not data.get("isGroup", False)
        ]

        user_info = [
            {
                "id": self.prefix + user_id,
                "login": user_id,
                "title": data.get("fullname", user_id),
                "description": data.get("fullname", user_id),
                "email": data.get("email", ""),
                "pluginid": plugin_id,
            }
            for (user_id, data) in users
        ]

        return tuple(user_info)

    def updateUser(self, user_id, login_name):
        """Update the login name of the user with id user_id.

        This is a new part of the IUserEnumerationPlugin interface, but
        not interesting for us.
        """
        pass

    def updateEveryLoginName(self, quit_on_first_error=True):
        """Update login names of all users to their canonical value.

        This is a new part of the IUserEnumerationPlugin interface, but
        not interesting for us.
        """
        pass


InitializeClass(ZODBMutablePropertyProvider)


class PersistentProperties(PersistentMapping):
    pass



from AccessControl import Permissions
from AccessControl import Unauthorized
from Acquisition import aq_parent
from plone.app.testing import logout
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING

import unittest


def sortTuple(t):
    l = list(t)
    l.sort()
    return tuple(l)


class TestGroupDataTool(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.acl_users = self.portal.acl_users
        self.groups = self.portal.portal_groups
        self.groupdata = self.portal.portal_groupdata
        self.groups.addGroup("foo")
        # MUST reset _v_ attributes!
        self.groupdata._v_temps = None
        if "auto_group" in self.acl_users:
            self.acl_users.manage_delObjects(["auto_group"])

    def testWrapGroup(self):
        g = self.acl_users.getGroup("foo")
        self.assertEqual(g.__class__.__name__, "PloneGroup")
        g = self.groupdata.wrapGroup(g)
        self.assertEqual(g.__class__.__name__, "GroupData")
        self.assertEqual(aq_parent(g).__class__.__name__, "PloneGroup")
        self.assertEqual(aq_parent(aq_parent(g)).__class__.__name__, "GroupManager")


class TestGroupData(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.membership = self.portal.portal_membership
        self.memberdata = self.portal.portal_memberdata
        self.acl_users = self.portal.acl_users
        self.groups = self.portal.portal_groups
        self.groupdata = self.portal.portal_groupdata
        self.groups.addGroup("foo")
        if "auto_group" in self.acl_users:
            self.acl_users.manage_delObjects(["auto_group"])
        # MUST reset _v_ attributes!
        self.memberdata._v_temps = None
        self.groupdata._v_temps = None

    def testGetGroup(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(g.__class__.__name__, "GroupData")
        g = g.getGroup()
        self.assertEqual(g.__class__.__name__, "PloneGroup")

    def testGetTool(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(g.getTool().getId(), "portal_groupdata")

    def testGetGroupMembers(self):
        g = self.groups.getGroupById("foo")
        self.acl_users.userSetGroups(TEST_USER_ID, groupnames=["foo"])
        self.assertEqual(g.getGroupMembers()[0].getId(), TEST_USER_ID)

    def testGroupMembersAreWrapped(self):
        g = self.groups.getGroupById("foo")
        self.acl_users.userSetGroups(TEST_USER_ID, groupnames=["foo"])
        ms = g.getGroupMembers()
        self.assertEqual(ms[0].__class__.__name__, "MemberData")
        self.assertEqual(aq_parent(ms[0]).__class__.__name__, "PluggableAuthService")

    def testAddMember(self):
        self.portal.manage_role("Member", [Permissions.manage_users])
        g = self.groups.getGroupById("foo")
        g.addMember(TEST_USER_ID)
        self.assertEqual(g.getGroupMembers()[0].getId(), TEST_USER_ID)

    def testRemoveMember(self):
        self.portal.manage_role("Member", [Permissions.manage_users])
        g = self.groups.getGroupById("foo")
        g.addMember(TEST_USER_ID)
        g.removeMember(TEST_USER_ID)
        self.assertEqual(len(g.getGroupMembers()), 0)

    def testSetGroupProperties(self):
        g = self.groups.getGroupById("foo")
        g.setGroupProperties({"email": "foo@bar.com"})
        gd = self.groups.getGroupById("foo")
        self.assertEqual(gd.getProperty("email"), "foo@bar.com")

    def testSetMemberProperties(self):
        # For reference
        m = self.membership.getMemberById(TEST_USER_ID)
        m.setMemberProperties({"email": "foo@bar.com"})
        md = self.membership.getMemberById(TEST_USER_ID)
        self.assertEqual(md.getProperty("email"), "foo@bar.com")

    def testGetProperty(self):
        g = self.groups.getGroupById("foo")
        g.setGroupProperties({"email": "foo@bar.com"})
        self.assertEqual(g.getProperty("email"), "foo@bar.com")
        self.assertEqual(g.getProperty("id"), "foo")

    def testGetGroupName(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(g.getGroupName(), "foo")

    def testGetGroupId(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(g.getGroupId(), "foo")

    def testGetRoles(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(tuple(g.getRoles()), ("Authenticated",))
        self.groups.editGroup(g.getId(), roles=["Member"])
        g = self.groups.getGroupById("foo")
        self.assertEqual(sortTuple(tuple(g.getRoles())), ("Authenticated", "Member"))

    def testGetRolesInContext(self):
        self.folder = self.portal["folder"]
        g = self.groups.getGroupById("foo")
        self.acl_users.userSetGroups(TEST_USER_ID, groupnames=["foo"])
        user = self.acl_users.getUser(TEST_USER_NAME)
        self.assertEqual(
            user.getRolesInContext(self.folder).sort(),
            ["Member", "Authenticated", "Owner"].sort(),
        )
        self.folder.manage_setLocalRoles(g.getId(), ["NewRole"])
        self.assertEqual(
            user.getRolesInContext(self.folder).sort(),
            ["Member", "Authenticated", "Owner", "NewRole"].sort(),
        )

    def testGetDomains(self):
        g = self.groups.getGroupById("foo")
        self.assertEqual(g.getDomains(), ())

    def testHasRole(self):
        g = self.groups.getGroupById("foo")
        self.groups.editGroup(g.getId(), roles=["Member"])
        g = self.groups.getGroupById("foo")
        self.assertTrue(g.has_role("Member"))


class TestMethodProtection(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.groups = self.portal.portal_groups
        self.groups.addGroup("foo")
        self.groupdata = self.groups.getGroupById("foo")

    def testAnonAddMember(self):
        logout()
        self.assertRaises(Unauthorized, self.groupdata.addMember, TEST_USER_ID)

    def testAnonRemoveMember(self):
        logout()
        self.assertRaises(Unauthorized, self.groupdata.removeMember, TEST_USER_ID)

    def testMemberAddMember(self):
        self.assertRaises(Unauthorized, self.groupdata.addMember, TEST_USER_ID)

    def testMemberRemoveMember(self):
        self.assertRaises(Unauthorized, self.groupdata.removeMember, TEST_USER_ID)

    def testManagerAddMember(self):
        self.portal.manage_role("Member", [Permissions.manage_users])
        self.groupdata.addMember(TEST_USER_ID)

    def testManagerRemoveMember(self):
        self.portal.manage_role("Member", [Permissions.manage_users])
        self.groupdata.addMember(TEST_USER_ID)
        self.groupdata.removeMember(TEST_USER_ID)



from Products.PluggableAuthService.interfaces import plugins
from zope.interface import Interface


class IGroupManagement(Interface):
    def addGroup(id, **kw):
        """
        Create a group with the supplied id, roles, and groups.
        return True if the operation suceeded
        """

    def addPrincipalToGroup(principal_id, group_id):
        """
        Add a given principal to the group.
        return True on success
        """

    def updateGroup(id, **kw):
        """
        Edit the given group. plugin specific
        return True on success
        """

    def setRolesForGroup(group_id, roles=()):
        """
        set roles for group
        return True on success
        """

    def removeGroup(group_id):
        """
        Remove the given group
        return True on success
        """

    def removePrincipalFromGroup(principal_id, group_id):
        """
        remove the given principal from the group
        return True on success
        """


class IGroupIntrospection(Interface):
    def getGroupById(group_id):
        """
        Returns the portal_groupdata-ish object for a group
        corresponding to this id.
        """

    #################################
    # these interface methods are suspect for scalability.
    #################################

    def getGroups():
        """
        Returns an iteration of the available groups
        """

    def getGroupIds():
        """
        Returns a list of the available groups
        """

    def getGroupMembers(group_id):
        """
        return the members of the given group
        """


class IGroupDataTool(Interface):
    def wrapGroup(group):
        """
        decorate a group with property management capabilities if needed
        """


class IGroupTool(IGroupIntrospection, IGroupManagement, plugins.IGroupsPlugin):

    """
    Defines an interface for managing and introspecting and
    groups and group membership.
    """


class IGroupData(Interface):
    """An abstract interface for accessing properties on a group object"""

    def setProperties(properties=None, **kw):
        """Allows setting of group properties en masse.
        Properties can be given either as a dict or a keyword parameters
        list"""

    def getProperty(id):
        """Return the value of the property specified by 'id'"""

    def getProperties():
        """Return the properties of this group. Properties are as usual in
        Zope."""

    def getGroupId():
        """Return the string id of this group, WITHOUT group prefix."""

    def getMemberId():
        """This exists only for a basic user/group API compatibility"""

    def getGroupName():
        """Return the name of the group."""

    def getGroupMembers():
        """Return a list of the portal_memberdata-ish members of the group."""

    def getAllGroupMembers():
        """Return a list of the portal_memberdata-ish members of the group
        including transitive ones (ie. users or groups of a group in that
        group)."""

    def getGroupMemberIds():
        """Return a list of the user ids of the group."""

    def getAllGroupMemberIds():
        """Return a list of the user ids of the group.
        including transitive ones (ie. users or groups of a group in that
        group)."""

    def addMember(id):
        """Add the existing member with the given id to the group"""

    def removeMember(id):
        """Remove the member with the provided id from the group"""

    def getGroup():
        """Returns the actual group implementation. Varies by group
        implementation (GRUF/Nux/et al)."""



# $Id$
"""Tests for Products.PlonePAS.plugins.role.GroupAwareRoleManager"""

from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING
from Products.PluggableAuthService.interfaces.plugins import IGroupsPlugin
from Products.PluggableAuthService.PluggableAuthService import _PLUGIN_TYPE_INFO
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.plugins.tests.helpers import DummyUser
from Products.PluggableAuthService.plugins.tests.helpers import FauxPAS
from Products.PluggableAuthService.plugins.tests.helpers import makeRequestAndResponse
from Products.PluginRegistry.PluginRegistry import PluginRegistry
from zope.interface import implementer

import unittest


@implementer(IGroupsPlugin)
class FauxGroupsPlugin(BasePlugin):
    def getGroupsForPrincipal(self, principal, request=None):
        return principal._groups


class GroupAwareRoleManagerTests(unittest.TestCase):
    """Roles manager that takes care of goup of principal"""

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def _getTargetClass(self):

        from Products.PlonePAS.plugins.role import GroupAwareRoleManager

        return GroupAwareRoleManager

    def _makeOne(self, id="test", *args, **kw):

        plugin = self._getTargetClass()(id=id, *args, **kw)
        # We need to bind a fake request to this plugin
        request, dummy_response = makeRequestAndResponse()
        setattr(plugin, "REQUEST", request)
        return plugin

    def test_roles_for_control_panel(self):
        """There's a special case, the users control panel for which
        we should never grant to users the roles they have got through
        the groups they belong.
        In that intent, the control panels view pushes
        '__ignore_group_roles__' = True
        in the request.
        """
        root = FauxPAS()

        # Add a minimal PluginRegistry with a mock IGroupsPlugin, because the
        # roles plugin depends on it:
        root._setObject("plugins", PluginRegistry(_PLUGIN_TYPE_INFO))
        root._setObject("groups", FauxGroupsPlugin())
        root["plugins"].activatePlugin(IGroupsPlugin, "groups")

        garm = self._makeOne("garm").__of__(root)

        # 2 roles
        garm.addRole("foo_role")
        garm.addRole("bar_role")

        # Group 'somegroup' has 'bar_role'
        garm.assignRoleToPrincipal("bar_role", "somegroup")

        # 'johndoe' has 'foo_role'
        johndoe = DummyUser("johndoe", ("somegroup",))
        garm.assignRoleToPrincipal("foo_role", "johndoe")

        # 'johndoe' should have 'foo_role' and 'bar_roles'
        got = garm.getRolesForPrincipal(johndoe)
        expected = ["foo_role", "bar_role"]
        self.assertEqual(set(got), set(expected))

        # For the users control panel, johndoe has only the 'foo_role'
        garm.REQUEST.set("__ignore_group_roles__", True)
        got = garm.getRolesForPrincipal(johndoe)
        self.assertEqual(got, ("foo_role",))

        # Confirm we can get only the inherited roles
        garm.REQUEST.set("__ignore_group_roles__", False)
        garm.REQUEST.set("__ignore_direct_roles__", True)
        got = garm.getRolesForPrincipal(johndoe)
        self.assertEqual(got, ("bar_role",))

        return



from Products.CMFCore import interfaces


class IMembershipTool(interfaces.IMembershipTool):
    def getMemberInfo(memberId=None):
        """Return 'harmless' Memberinfo of any member, such as full name,
        location, etc
        """


__all__ = ("IMembershipTool",)



from DateTime import DateTime
from OFS.Image import Image
from plone.app.testing import TEST_USER_ID as default_user
from Products.CMFCore.interfaces import IMember
from Products.CMFCore.interfaces import IMemberData
from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING
from Products.PlonePAS.tests import dummy
from Products.PluggableAuthService.interfaces.events import IPropertiesUpdatedEvent
from zope.component import getGlobalSiteManager
from zope.component import getMultiAdapter

import unittest


class TestMemberDataTool(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.memberdata = self.portal.portal_memberdata
        self.membership = self.portal.portal_membership
        self.membership.memberareaCreationFlag = 0
        # Don't let default_user disturb results
        self.portal.acl_users._doDelUsers([default_user])
        # Add some members
        self.addMember(
            "fred",
            "Fred Flintstone",
            "fred@bedrock.com",
            ["Member", "Reviewer"],
            "2002-01-01",
        )
        self.addMember(
            "barney", "Barney Rubble", "barney@bedrock.com", ["Member"], "2002-01-01"
        )
        self.addMember(
            "brubble", "Bambam Rubble", "bambam@bambam.net", ["Member"], "2003-12-31"
        )
        # MUST reset this
        self.memberdata._v_temps = None

    def addMember(self, username, fullname, email, roles, last_login_time):
        self.membership.addMember(username, "secret", roles, [])
        member = self.membership.getMemberById(username)
        member.setMemberProperties(
            {
                "fullname": fullname,
                "email": email,
                "last_login_time": DateTime(last_login_time),
            }
        )

    def testSetPortrait(self):
        self.memberdata._setPortrait(
            Image(id=default_user, file=dummy.File(), title=""), default_user
        )
        self.assertEqual(
            self.memberdata._getPortrait(default_user).getId(), default_user
        )
        self.assertEqual(self.memberdata._getPortrait(default_user).meta_type, "Image")

    def testDeletePortrait(self):
        self.memberdata._setPortrait(
            Image(id=default_user, file=dummy.File(), title=""), default_user
        )
        self.memberdata._deletePortrait(default_user)
        self.assertEqual(self.memberdata._getPortrait(default_user), None)

    def testPruneMemberDataContents(self):
        # Only test what is not already tested elswhere
        self.memberdata._setPortrait(
            Image(id=default_user, file=dummy.File(), title=""), default_user
        )
        self.memberdata._setPortrait(
            Image(id=default_user, file=dummy.File(), title=""), "dummy"
        )
        self.memberdata.pruneMemberDataContents()
        self.assertEqual(len(self.memberdata.portraits), 1)

    def testFulltextMemberSearch(self):
        # Search for a user by id, name, email, ...
        search = self.memberdata.searchFulltextForMembers
        self.assertEqual(len(search("")), 3)
        self.assertEqual(len(search("rubble")), 2)
        self.assertEqual(len(search("stone")), 1)
        self.assertEqual(len(search("bambam.net")), 1)
        self.assertEqual(len(search("bedrock.com")), 2)
        self.assertEqual(len(search("brubble")), 1)

    def testMemberDataAdapter(self):
        """Test, if the PlonePAS MemberData adapter is used instead of the
        default one from Products.CMFCore.MemberDataTool
        """
        from Products.PlonePAS.tools.memberdata import MemberData

        member = self.membership.getMemberById("fred")

        adapter = getMultiAdapter((member, self.memberdata), IMember)
        self.assertEqual(adapter.__class__, MemberData)

        wrapped_user = self.memberdata.wrapUser(member)
        self.assertEqual(wrapped_user.__class__, MemberData)

    def testPropertiesUpdatedEvent(self):
        def event_handler(context, event):
            self._properties_updated_handler_called = True

        gsm = getGlobalSiteManager()
        gsm.registerHandler(event_handler, (IMemberData, IPropertiesUpdatedEvent))

        self._properties_updated_handler_called = False

        username = "ez"
        roles = ["Member"]
        fullname = "Ez Zy"
        email = "ez@ezmail.net"

        self.membership.addMember(username, "secret", roles, [])
        member = self.membership.getMemberById(username)

        self.assertFalse(self._properties_updated_handler_called)

        member.setMemberProperties({"fullname": fullname, "email": email})

        self.assertTrue(self._properties_updated_handler_called)

        # Test that notify(PropertiesUpdated) isn't called on user login.
        self._properties_updated_handler_called = False

        # Imitate a login as the plone.app.testing login method doesn't seem to
        # set these member properties.
        member.setMemberProperties(
            {
                "login_time": DateTime("2018-02-15"),
                "last_login_time": DateTime("2018-02-15"),
            }
        )

        self.assertFalse(self._properties_updated_handler_called)

        # Test notify(PropertiesUpdated) isn't called when login_time is
        # present as we're assuming this should only be changed on login.
        self._properties_updated_handler_called = False
        member.setMemberProperties(
            {"login_time": DateTime("2018-02-15"), "fullname": "Bed Rock"}
        )

        self.assertFalse(self._properties_updated_handler_called)
        gsm.unregisterHandler(event_handler, (IMemberData, IPropertiesUpdatedEvent))



from io import BytesIO
from OFS.SimpleItem import SimpleItem
from ZPublisher.HTTPRequest import FileUpload

import typing

TEXT = b"file data"


class FieldStorage:
    def __init__(self, file, filename="testfile", headers=None):
        self.file = file
        if headers is None:
            headers = {}
        self.headers = headers
        self.filename = filename
        self.name = filename


class File(FileUpload):
    """Dummy upload object
    Used to fake uploaded files.
    """

    __allow_access_to_unprotected_subobjects__ = 1
    filename = "dummy.txt"
    data = TEXT
    headers: typing.Dict = {}

    def __init__(self, filename=None, data=None, headers=None):
        if filename is not None:
            self.filename = filename
        if data is not None:
            self.data = data
        if headers is not None:
            self.headers = headers
        self.file = BytesIO(self.data)

    def seek(self, *args):
        pass

    def tell(self, *args):
        return 1

    def read(self, *args):
        return self.data


class Error(Exception):
    """Dummy exception"""


class Raiser(SimpleItem):
    """Raises the stored exception when called"""

    exception = Error

    def __init__(self, exception=None):
        if exception is not None:
            self.exception = exception

    def __call__(self, *args, **kw):
        raise self.exception



from plone.memoize.instance import memoize
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.interfaces.browser import IPASMemberView
from zope.interface import implementer
from zope.publisher.browser import BrowserView


@implementer(IPASMemberView)
class PASMemberView(BrowserView):
    @memoize
    def info(self, userid=None):
        pm = getToolByName(self.context, "portal_membership")
        result = pm.getMemberInfo(memberId=userid)
        if result is None:
            # No such member: removed?  We return something useful anyway.
            return {
                "username": userid,
                "description": "",
                "language": "",
                "home_page": "",
                "name_or_id": userid,
                "location": "",
                "fullname": "",
            }
        result["name_or_id"] = (
            result.get("fullname") or result.get("username") or userid
        )
        return result



from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.CMFCore.utils import getToolByName
from Products.PlonePAS.plugins.property import ZODBMutablePropertyProvider
from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING
from Products.PluggableAuthService.interfaces.plugins import IUserEnumerationPlugin

import unittest


class PropertiesTest(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]

    def test_user_properties(self):
        mt = getToolByName(self.portal, "portal_membership")
        md = getToolByName(self.portal, "portal_memberdata")

        # Create a new Member
        mt.addMember(
            "user1",
            "u1",
            ["Member"],
            [],
            {"email": "user1@host.com", "fullname": "User #1"},
        )
        member = mt.getMemberById("user1")
        self.assertFalse(member is None)

        # Assert user doesn't have the property yet
        self.assertFalse(member.hasProperty("age"))

        # Add new property
        md.manage_addProperty("age", 20, "int")

        # Assert user has the property now
        self.assertTrue(member.hasProperty("age"))

        # Get the property, should have the default value
        got = member.getProperty("age", None)
        expected = 20
        self.assertEqual(got, expected)

        # get a handle on the member
        member = mt.getMemberById("user1")

        # Set some member properties. Needs to be logged in as the user.
        login(self.portal, "user1")
        member.setMemberProperties(
            {"age": 30, "fullname": "User #1 Is Cool", "email": "user1@anotherhost.qa"}
        )

        # Check the properties have been set
        got = member.getProperty("age", None)
        expected = 30
        self.assertEqual(got, expected)

        got = member.getProperty("fullname", None)
        expected = "User #1 Is Cool"
        self.assertEqual(got, expected)

        got = member.getProperty("email", None)
        expected = "user1@anotherhost.qa"
        self.assertEqual(got, expected)

        # Delete the property
        md.manage_delProperties(ids=("age",))

        # re-get the member to reflect the new memberdata schema
        member = mt.getMemberById("user1")

        # Assert property is gone
        self.assertFalse(member.hasProperty("age"))

        # Get the property, should return default (None)
        got = member.getProperty("age", None)
        expected = None
        self.assertEqual(got, expected)

        # Other properties should still be there.
        got = member.getProperty("fullname", None)
        expected = "User #1 Is Cool"
        self.assertEqual(got, expected)

        got = member.getProperty("email", None)
        expected = "user1@anotherhost.qa"
        self.assertEqual(got, expected)

    def test_group_properties(self):
        gt = getToolByName(self.portal, "portal_groups")
        gd = getToolByName(self.portal, "portal_groupdata")

        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        # Create a new Group
        gt.addGroup(
            "group1",
            ["Reviewer"],
            [],
            {"email": "group1@host.com", "title": "Group #1"},
        )
        group = gt.getGroupById("group1")
        self.assertFalse(group is None)

        # Assert group doesn't have the property yet
        self.assertFalse(group.hasProperty("karma"))

        # Add new property
        gd.manage_addProperty("karma", 20, "int")

        # get group again to re-create with new groupdata schema
        group = gt.getGroupById("group1")

        # Assert group has the property now
        self.assertTrue(group.hasProperty("karma"))

        # Get the property, should have the default value
        got = group.getProperty("karma", None)
        expected = 20

        self.assertEqual(got, expected)

        # Set some group properties
        group.setGroupProperties(
            {"karma": 30, "title": "Group #1 Is Cool", "email": "group1@anotherhost.qa"}
        )

        # Check the properties have been set
        got = group.getProperty("karma", None)
        expected = 30
        self.assertEqual(got, expected)

        got = group.getProperty("title", None)
        expected = "Group #1 Is Cool"
        self.assertEqual(got, expected)

        got = group.getProperty("email", None)
        expected = "group1@anotherhost.qa"
        self.assertEqual(got, expected)

        # Delete the property
        gd.manage_delProperties(ids=("karma",))

        # get group again to re-create with new groupdata schema
        group = gt.getGroupById("group1")

        # Assert property is gone
        self.assertFalse(group.hasProperty("karma"))

        # Get the property, should return default (None)
        got = group.getProperty("karma", None)
        expected = None
        self.assertEqual(got, expected)

        # Other properties should still be there.
        got = group.getProperty("title", None)
        expected = "Group #1 Is Cool"
        self.assertEqual(got, expected)

        got = group.getProperty("email", None)
        expected = "group1@anotherhost.qa"
        self.assertEqual(got, expected)

    def test_schema_for_mutable_property_provider(self):
        """Add a schema to a ZODBMutablePropertyProvider."""

        # Schema is list of tuples with name, type (string), value.
        # From the types it seems only 'lines' is handled differently.
        address_schema = [
            ("addresses", "lines", ["Here", "There"]),
            ("city", "str", "Somewhere"),
            ("telephone", "int", 1234567),
        ]

        # This used to give a ValueError, so we just check that it
        # does not.
        provider = ZODBMutablePropertyProvider(
            "address_plugin", "Address Plugin", schema=address_schema
        )

        # When this test passes, we are happy already, but let's add a
        # few more basic tests.

        # Create a new Member
        mt = getToolByName(self.portal, "portal_membership")
        mt.addMember(
            "user1",
            "u1",
            ["Member"],
            [],
            {"email": "user1@host.com", "fullname": "User #1"},
        )
        member = mt.getMemberById("user1")
        sheet = provider.getPropertiesForUser(member)
        self.assertEqual(sheet.propertyIds(), ["addresses", "city", "telephone"])
        self.assertEqual(
            sheet.propertyInfo("city"), {"type": "str", "id": "city", "mode": ""}
        )
        self.assertEqual(sheet.getProperty("addresses"), ("Here", "There"))


class PropertySearchTest(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.mt = getToolByName(self.portal, "portal_membership")
        self.md = getToolByName(self.portal, "portal_memberdata")
        self.gt = getToolByName(self.portal, "portal_groups")

        # Create a new Member
        self.mt.addMember(
            "member1",
            "pw",
            ["Member"],
            [],
            {"email": "member1@host.com", "title": "Member #1"},
        )
        member = self.mt.getMemberById("member1")
        self.assertFalse(member is None)

        self.mt.addMember(
            "member2",
            "pw",
            ["Member"],
            [],
            {"email": "user2@otherhost.com", "fullname": "User #2"},
        )
        member = self.mt.getMemberById("member2")
        self.assertFalse(member is None)

        # Add a Group to make sure searchUsers isn't returning them in results.
        self.gt.addGroup("group1", title="Group 1")
        group = self.gt.getGroupById("group1")
        self.assertFalse(group is None)

        self.pas = getToolByName(self.portal, "acl_users")
        for plugin in self.pas.plugins.getAllPlugins("IUserEnumerationPlugin")[
            "active"
        ]:
            if plugin != "mutable_properties":
                self.pas.plugins.deactivatePlugin(IUserEnumerationPlugin, plugin)

    def testPluginActivated(self):
        plugins = self.pas.plugins.getAllPlugins("IUserEnumerationPlugin")["active"]
        self.assertEqual(plugins, ("mutable_properties",))

    def testEmptySearch(self):
        results = self.pas.searchUsers()
        self.assertEqual(len(results), 2)

    def testInexactStringSearch(self):
        results = self.pas.searchUsers(email="something@somewhere.tld")
        self.assertEqual(results, ())

        results = self.pas.searchUsers(email="member1@host.com", exact_match=False)
        results = [info["userid"] for info in results]
        self.assertEqual(results, ["member1"])

        results = self.pas.searchUsers(email="@host.com", exact_match=False)
        results = [info["userid"] for info in results]
        self.assertEqual(results, ["member1"])

        results = self.pas.searchUsers(email="member1@host.com", exact_match=True)
        results = [info["userid"] for info in results]
        self.assertEqual(results, ["member1"])

        results = self.pas.searchUsers(email="@host.com", exact_match=True)
        results = [info["userid"] for info in results]
        self.assertEqual(results, [])

    def testBooleanSearch(self):
        results = self.pas.searchUsers(visible_ids=True)
        results = [info["userid"] for info in results]
        self.assertEqual(results, [])

        results = self.pas.searchUsers(visible_ids=False)
        results = [info["userid"] for info in results]
        self.assertEqual(results, ["member1", "member2"])

    def testGroupsNotReturnedByEnumerateUsers(self):
        """Check to make sure that groups aren't returned by a enumerateUsers
        call.
        See http://dev.plone.org/plone/ticket/9435"""
        results = self.pas.searchUsers()
        resultIds = [a["id"] for a in results]
        self.assertFalse("group1" in resultIds)

    def testSearchEmptyId(self):
        self.assertEqual(self.pas.mutable_properties.enumerateUsers(id=""), ())
        self.assertEqual(self.pas.mutable_properties.enumerateUsers(login=""), ())

    def testCantSearchByIdOrLogin(self):
        # we can't search by id
        results = self.pas.searchUsers(id="member1")
        self.assertEqual(results, ())
        # or login
        results = self.pas.searchUsers(login="member1")
        self.assertEqual(results, ())



from Products.PlonePAS.patch import call
from Products.PlonePAS.patch import wrap_method
from Products.PlonePAS.plugins.group import PloneGroup


GROUP_PROPERTY_MAP = {
    # target property: (possible key, ...)
    "title": (
        "name",
        "displayName",
        "cn",
    ),
    "description": ("description",),
    "email": ("mail",),
}

KNOWN_ATTRS = set()
for attrs in GROUP_PROPERTY_MAP.values():
    for attr in attrs:
        KNOWN_ATTRS.add(attr)


def getPropertiesForUser(self, user, request=None):
    """Fullfill PropertiesPlugin requirements"""

    if not isinstance(user, PloneGroup):
        # It's not a PloneGroup, continue as usual
        return call(self, "getPropertiesForUser", user=user, request=request)

    # We've got a PloneGroup.
    # Search for groups instead of users
    groups = self.enumerateGroups(id=user.getId(), exact_match=True)
    # XXX Should we assert there's only one group?
    properties = {}
    for group in groups:
        for pname, attrs in GROUP_PROPERTY_MAP.items():
            for attr in attrs:
                value = group.get(attr)
                if value:
                    # Break on first found
                    properties[pname] = value
                    break

    return properties


def getGroupsForPrincipal(self, user, request=None, attr=None):
    """Fulfill GroupsPlugin requirements, but don't return any groups for
    groups"""

    if not isinstance(user, PloneGroup):
        # It's not a PloneGroup, continue as usual
        return call(self, "getGroupsForPrincipal", user, request=request, attr=attr)

    return ()


def patch_ldapmp():
    from Products.LDAPMultiPlugins.LDAPPluginBase import LDAPPluginBase

    wrap_method(LDAPPluginBase, "getPropertiesForUser", getPropertiesForUser)

    from Products.LDAPMultiPlugins.LDAPMultiPlugin import LDAPMultiPlugin

    wrap_method(LDAPMultiPlugin, "getGroupsForPrincipal", getGroupsForPrincipal)



"""
Custom GenericSetup import steps for PAS in Plone.
"""

from Acquisition import aq_base
from Acquisition import aq_parent
from plone.session.plugins.session import manage_addSessionPlugin
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import interfaces as plone_ifaces
from Products.PlonePAS import config
from Products.PlonePAS.interfaces import group as igroup
from Products.PlonePAS.interfaces.plugins import ILocalRolesPlugin
from Products.PlonePAS.interfaces.plugins import IUserIntrospection
from Products.PlonePAS.interfaces.plugins import IUserManagement
from Products.PlonePAS.plugins import cookie_handler
from Products.PluggableAuthService.Extensions.upgrade import replace_acl_users
from Products.PluggableAuthService.interfaces.authservice import IPluggableAuthService
from Products.PluggableAuthService.interfaces.plugins import IChallengePlugin
from Products.PluggableAuthService.interfaces.plugins import ICredentialsResetPlugin
from Products.PluggableAuthService.plugins import CookieAuthHelper
from Products.PluggableAuthService.plugins.RecursiveGroupsPlugin import (
    addRecursiveGroupsPlugin,
)
from zope import component

import logging


logger = logging.getLogger("PlonePAS setup")


def activatePluginInterfaces(portal, plugin, disable=None):
    if disable is None:
        disable = []
    pas = portal.acl_users
    plugin_obj = pas[plugin]

    activatable = []

    for info in plugin_obj.plugins.listPluginTypeInfo():
        interface = info["interface"]
        interface_name = info["id"]
        if plugin_obj.testImplements(interface):
            if interface_name in disable:
                disable.append(interface_name)
                logger.debug("Disabling: " + info["title"])
            else:
                activatable.append(interface_name)
                logger.debug("Activating: " + info["title"])
    plugin_obj.manage_activateInterfaces(activatable)
    logger.debug(plugin + " activated.")


def setupRoles(portal):
    rmanager = portal.acl_users.role_manager
    rmanager.addRole("Member", title="Portal Member")
    rmanager.addRole("Reviewer", title="Content Reviewer")


def registerPluginType(pas, plugin_type, plugin_info):
    # Make sure there's no dupes in _plugin_types, otherwise your PAS
    # will *CRAWL*
    plugin_types = list(set(pas.plugins._plugin_types))
    if plugin_type not in plugin_types:
        plugin_types.append(plugin_type)

    # Order doesn't seem to matter, but let's store it ordered.
    plugin_types.sort()

    # Re-assign to the object, because this is a non-persistent list.
    pas.plugins._plugin_types = plugin_types

    # It's safe to assign over a existing key here.
    pas.plugins._plugin_type_info[plugin_type] = plugin_info


def registerPluginTypes(pas):

    PluginInfo = {
        "id": "IUserManagement",
        "title": "user_management",
        "description": (
            "The User Management plugins allow the "
            "Pluggable Auth Service to add/delete/modify users"
        ),
    }

    registerPluginType(pas, IUserManagement, PluginInfo)

    PluginInfo = {
        "id": "IUserIntrospection",
        "title": "user_introspection",
        "description": (
            "The User Introspection plugins allow the "
            "Pluggable Auth Service to provide lists of users"
        ),
    }

    registerPluginType(pas, IUserIntrospection, PluginInfo)

    PluginInfo = {
        "id": "IGroupManagement",
        "title": "group_management",
        "description": (
            "Group Management provides add/write/deletion "
            "of groups and member management"
        ),
    }

    registerPluginType(pas, igroup.IGroupManagement, PluginInfo)

    PluginInfo = {
        "id": "IGroupIntrospection",
        "title": "group_introspection",
        "description": (
            "Group Introspection provides listings " "of groups and membership"
        ),
    }

    registerPluginType(pas, igroup.IGroupIntrospection, PluginInfo)

    PluginInfo = {
        "id": "ILocalRolesPlugin",
        "title": "local_roles",
        "description": "Defines Policy for getting Local Roles",
    }

    registerPluginType(pas, ILocalRolesPlugin, PluginInfo)


def setupPlugins(portal):
    uf = portal.acl_users
    logger.debug("\nPlugin setup")

    pas = uf.manage_addProduct["PluggableAuthService"]
    plone_pas = uf.manage_addProduct["PlonePAS"]

    setupAuthPlugins(portal, pas, plone_pas)

    found = uf.objectIds(["User Manager"])
    if not found:
        plone_pas.manage_addUserManager("source_users")
        logger.debug("Added User Manager.")
    activatePluginInterfaces(portal, "source_users")

    found = uf.objectIds(["Group Aware Role Manager"])
    if not found:
        plone_pas.manage_addGroupAwareRoleManager("portal_role_manager")
        logger.debug("Added Group Aware Role Manager.")
        activatePluginInterfaces(portal, "portal_role_manager")

    found = uf.objectIds(["Local Roles Manager"])
    if not found:
        plone_pas.manage_addLocalRolesManager("local_roles")
        logger.debug("Added Group Aware Role Manager.")
        activatePluginInterfaces(portal, "local_roles")

    found = uf.objectIds(["Group Manager"])
    if not found:
        plone_pas.manage_addGroupManager("source_groups")
        logger.debug("Added ZODB Group Manager.")
        activatePluginInterfaces(portal, "source_groups")

    found = uf.objectIds(["Plone User Factory"])
    if not found:
        plone_pas.manage_addPloneUserFactory("user_factory")
        logger.debug("Added Plone User Factory.")
        activatePluginInterfaces(portal, "user_factory")

    found = uf.objectIds(["ZODB Mutable Property Provider"])
    if not found:
        plone_pas.manage_addZODBMutablePropertyProvider("mutable_properties")
        logger.debug("Added Mutable Property Manager.")
        activatePluginInterfaces(portal, "mutable_properties")

    found = uf.objectIds(["Automatic Group Plugin"])
    if not found:
        plone_pas.manage_addAutoGroup(
            "auto_group",
            "Authenticated Users (Virtual Group)",
            "AuthenticatedUsers",
            "Automatic Group Provider",
        )
        logger.debug("Added Automatic Group.")
        activatePluginInterfaces(portal, "auto_group")

    found = uf.objectIds(["Plone Session Plugin"])
    if not found:
        manage_addSessionPlugin(plone_pas, "session")
        logger.debug("Added Plone Session Plugin.")
        activatePluginInterfaces(portal, "session")

    found = uf.objectIds(["Recursive Groups Plugin"])
    if not found:
        addRecursiveGroupsPlugin(
            plone_pas, "recursive_groups", "Recursive Groups Plugin"
        )
        activatePluginInterfaces(portal, "recursive_groups")
        logger.debug("Added Recursive Groups plugin.")

    setupPasswordPolicyPlugin(portal)


def setupAuthPlugins(
    portal,
    pas,
    plone_pas,
    deactivate_basic_reset=True,
    deactivate_cookie_challenge=False,
):
    uf = portal.acl_users
    logger.debug("Cookie plugin setup")

    login_path = "login_form"
    cookie_name = "__ac"

    crumbler = getToolByName(portal, "cookie_authentication", None)
    if crumbler is not None:
        login_path = crumbler.auto_login_page
        cookie_name = crumbler.auth_cookie

    is_plone_site = plone_ifaces.IPloneSiteRoot.providedBy(portal)
    if is_plone_site:
        cookie_meta_type = cookie_handler.ExtendedCookieAuthHelper.meta_type
        add_cookie_plugin = plone_pas.manage_addExtendedCookieAuthHelper
    else:
        # Can't use the `ExtendedCookieAuthHelper` outside of a Plone portal.
        cookie_meta_type = CookieAuthHelper.CookieAuthHelper.meta_type
        add_cookie_plugin = pas.addCookieAuthHelper
    cookie_auth_ids = uf.objectIds(cookie_meta_type)
    if not cookie_auth_ids:
        add_cookie_plugin(
            "credentials_cookie_auth",
            cookie_name=cookie_name,
        )
        logger.debug(
            "Added %r: %r",
            cookie_meta_type,
            "/".join(uf.credentials_cookie_auth.getPhysicalPath()),
        )
    if deactivate_basic_reset:
        disable = ["ICredentialsResetPlugin", "ICredentialsUpdatePlugin"]
    else:
        disable = []
    activatePluginInterfaces(portal, "credentials_cookie_auth", disable=disable)

    credentials_cookie_auth = uf._getOb("credentials_cookie_auth")
    if is_plone_site and "login_form" in credentials_cookie_auth:
        credentials_cookie_auth.manage_delObjects(ids=["login_form"])
        logger.debug("Removed default login_form from credentials cookie " "auth.")
    credentials_cookie_auth.cookie_name = cookie_name
    credentials_cookie_auth.login_path = login_path

    # remove cookie crumbler(s)
    if "cookie_authentication" in portal:
        portal.manage_delObjects(["cookie_authentication"])
    logger.debug("Removed old Cookie Crumbler")

    found = uf.objectIds(["HTTP Basic Auth Helper"])
    if not found:
        pas.addHTTPBasicAuthHelper("credentials_basic_auth", title="HTTP Basic Auth")
    logger.debug("Added Basic Auth Helper.")
    activatePluginInterfaces(portal, "credentials_basic_auth")

    if deactivate_basic_reset:
        uf.plugins.deactivatePlugin(ICredentialsResetPlugin, "credentials_basic_auth")
    if deactivate_cookie_challenge:
        uf.plugins.deactivatePlugin(IChallengePlugin, "credentials_cookie_auth")


def updateProperties(tool, properties):
    dependency_keys = ("selection", "multiple selection")
    propsWithNoDeps = [
        prop for prop in properties if prop["type"] not in dependency_keys
    ]
    propsWithDeps = [prop for prop in properties if prop["type"] in dependency_keys]
    for prop in propsWithNoDeps:
        updateProp(tool, prop)
    for prop in propsWithDeps:
        updateProp(tool, prop)


def updateProp(prop_manager, prop_dict):
    """Provided a PropertyManager and a property dict of {id, value,
    type}, set or update that property as applicable.

    Doesn't deal with existing properties changing type.
    """
    id = prop_dict["id"]
    value = prop_dict["value"]
    type = prop_dict["type"]
    if type in ("selection", "multiple selection"):
        value = prop_dict["select_variable"]
    if prop_manager.hasProperty(id):
        prop_manager._updateProperty(id, value)
    else:
        prop_manager._setProperty(id, value, type)
    if type in ("selection", "multiple selection"):
        prop_manager._updateProperty(id, prop_dict["value"])


def addPAS(portal):
    logger.debug("Adding PAS user folder")
    portal.manage_addProduct["PluggableAuthService"].addPluggableAuthService()


def migrate_root_uf(self):
    # Acquire parent user folder.
    parent = self.getPhysicalRoot()
    uf = getToolByName(parent, "acl_users")
    if IPluggableAuthService.providedBy(uf):
        # It's a PAS already, fixup if needed.
        pas_fixup(parent)

        # Configure Challenge Chooser plugin if available
        challenge_chooser_setup(parent)
        return

    if not uf.meta_type == "User Folder":
        # It's not a standard User Folder at the root. Nothing we can do.
        return

    # It's a standard User Folder, replace it.
    replace_acl_users(parent)

    # Get the new uf
    uf = getToolByName(parent, "acl_users")

    pas = uf.manage_addProduct["PluggableAuthService"]
    plone_pas = uf.manage_addProduct["PlonePAS"]
    # Setup authentication plugins
    setupAuthPlugins(
        parent,
        pas,
        plone_pas,
        deactivate_basic_reset=False,
        deactivate_cookie_challenge=True,
    )

    # Activate *all* interfaces for user manager. IUserAdder is not
    # activated for some reason by default.
    activatePluginInterfaces(parent, "users")

    # Configure Challenge Chooser plugin if available
    challenge_chooser_setup(parent)


def pas_fixup(self):
    from Products.PluggableAuthService.PluggableAuthService import _PLUGIN_TYPE_INFO

    pas = getToolByName(self, "acl_users")
    if not IPluggableAuthService.providedBy(pas):
        logger.debug("PAS UF not found, skipping PAS fixup.")
        return

    plugins = pas["plugins"]

    plugin_types = list(set(plugins._plugin_types))
    for key, id, title, description in _PLUGIN_TYPE_INFO:
        if key in plugin_types:
            logger.debug("Plugin type '%s' already registered." % id)
            continue
        logger.debug("Plugin type '%s' was not registered." % id)
        plugin_types.append(key)
        plugins._plugin_type_info[key] = {
            "id": id,
            "title": title,
            "description": description,
        }
    # Make it ordered
    plugin_types.sort()

    # Re-assign because it's a non-persistent property.
    plugins._plugin_types = plugin_types


def challenge_chooser_setup(self):
    uf = getToolByName(self, "acl_users")
    pas = uf.manage_addProduct["PluggableAuthService"]

    # Only install plugins if available
    req = ("addChallengeProtocolChooserPlugin", "addRequestTypeSnifferPlugin")
    for m in req:
        if getattr(pas, m, None) is None:
            logger.debug("Needed plugins have not been found, ignoring")
            return

    found = uf.objectIds(["Challenge Protocol Chooser Plugin"])
    if not found:
        logger.debug("Adding Challenge Protocol Chooser Plugin.")
        pas.addChallengeProtocolChooserPlugin(
            "chooser", mapping=config.DEFAULT_PROTO_MAPPING
        )
        activatePluginInterfaces(self, "chooser")
    else:
        assert len(found) == 1, "Found extra plugins %s" % found
        logger.debug("Found existing Challenge Protocol Chooser Plugin.")
        plugin = uf[found[0]]
        plugin.manage_updateProtocolMapping(mapping=config.DEFAULT_PROTO_MAPPING)
        activatePluginInterfaces(self, found[0])

    found = uf.objectIds(["Request Type Sniffer Plugin"])
    if not found:
        logger.debug("Adding Request Type Sniffer Plugin.")
        pas.addRequestTypeSnifferPlugin("sniffer")
        activatePluginInterfaces(self, "sniffer")
    else:
        assert len(found) == 1, "Found extra plugins %s" % found
        logger.debug("Found existing Request Type Sniffer Plugin.")
        activatePluginInterfaces(self, found[0])


def setupPasswordPolicyPlugin(portal):
    uf = portal.acl_users
    plone_pas = uf.manage_addProduct["PlonePAS"]

    found = uf.objectIds(["Default Plone Password Policy"])
    logger.debug("\nDefault Password Ploicy Plugin setup")
    if not found:
        plone_pas.manage_addPasswordPolicyPlugin(
            "password_policy", title="Default Plone Password Policy"
        )
        logger.debug("Added Default Plone Password Policy.")
        activatePluginInterfaces(portal, "password_policy")


def setLoginFormInCookieAuth(context):
    """Makes sure the cookie auth redirects to 'require_login' instead
    of 'login_form'."""
    uf = getattr(context, "acl_users", None)
    if uf is None or getattr(aq_base(uf), "_getOb", None) is None:
        # we have no user folder or it's not a PAS folder, do nothing
        return
    cookie_auth = uf._getOb("credentials_cookie_auth", None)
    if cookie_auth is None:
        # there's no cookie auth object, do nothing
        return
    current_login_form = cookie_auth.getProperty("login_path")
    if current_login_form != "login_form":
        # it's customized already, do nothing
        return
    cookie_auth.manage_changeProperties(login_path="require_login")


def addRolesToPlugIn(p):
    """
    XXX This is horrible.. need to switch PlonePAS to a GenericSetup
    based install so this doesn't need to happen.

    Have to manually register the roles from the 'rolemap' step
    with the roles plug-in.
    """
    uf = getToolByName(p, "acl_users")
    rmanager = uf.portal_role_manager
    roles = ("Reviewer", "Member")
    existing = rmanager.listRoleIds()
    for role in roles:
        if role not in existing:
            rmanager.addRole(role)


def setupGroups(site):
    """
    Create Plone's default set of groups.
    """
    uf = getToolByName(site, "acl_users")
    gtool = getToolByName(site, "portal_groups")
    if not uf.searchGroups(id="Administrators"):
        gtool.addGroup("Administrators", title="Administrators", roles=["Manager"])

    if not uf.searchGroups(id="Site Administrators"):
        gtool.addGroup(
            "Site Administrators",
            title="Site Administrators",
            roles=["Site Administrator"],
        )

    if not uf.searchGroups(id="Reviewers"):
        gtool.addGroup("Reviewers", title="Reviewers", roles=["Reviewer"])


def installPAS(portal):
    # Add user folder
    portal.manage_addProduct["PluggableAuthService"].addPluggableAuthService()

    # Configure Challenge Chooser plugin if available
    challenge_chooser_setup(portal)

    # A bunch of general configuration settings
    registerPluginTypes(portal.acl_users)
    setupPlugins(portal)

    # XXX Why are we doing this?
    # According to Sidnei, "either cookie or basic auth for a user in the root
    # folder doesn't work
    # if it's not a PAS UF when you sign in to Plone. IIRC."
    # See: http://twitter.com/#!/sidneidasilva/status/14030732112429056
    # And here's the original commit:
    # http://dev.plone.org/collective/changeset/10720/PlonePAS/trunk/Extensions/Install.py
    if aq_parent(portal):
        migrate_root_uf(portal)


def setupPlonePAS(context):
    """
    Setup PlonePAS step.
    """
    # Only run step if a flag file is present (e.g. not an extension profile)
    if context.readDataFile("plone-pas.txt") is None:
        return
    site = context.getSite()
    if "acl_users" not in site:
        installPAS(site)
        addRolesToPlugIn(site)
        setupGroups(site)
        setLoginFormInCookieAuth(site)


def set_up_zope_root_cookie_auth(context):
    """
    Change the Zope root `/acl_users` to use a simple cookie login form.
    """
    portal = aq_parent(context)
    root = portal.getPhysicalRoot()
    root_acl_users = getToolByName(root, "acl_users")

    # Enable the cookie plugin for all interfaces
    activatePluginInterfaces(root, "credentials_cookie_auth")
    # Ensure that the cookie login form is used to challenge for authentication
    credentials_cookie_auth = root_acl_users._getOb(  # pylint: disable=protected-access
        "credentials_cookie_auth",
    )
    root_acl_users.plugins.movePluginsTop(
        IChallengePlugin,
        [credentials_cookie_auth.id],
    )
    # Disable the HTTP `Basic ...` authentication plugin
    for plugin_iface in (
        IChallengePlugin,
        # Apparently, the `HTTPBasicAuthHelper` plugin"s `ICredentialsResetPlugin`
        # implementation interferes with deleting/expiring cookies, specifically the
        # `__ac` cookie in this case.  I first tried moving that plugin to the top and
        # bottom for that interface, but the cookies still remained after logout.  Only
        # deactivating it worked.
        ICredentialsResetPlugin,
    ):
        activated_plugin_ids = root_acl_users.plugins.listPluginIds(plugin_iface)
        if "credentials_basic_auth" in activated_plugin_ids:
            root_acl_users.plugins.deactivatePlugin(
                plugin_iface,
                "credentials_basic_auth",
            )



"""
Add Mutable Property Sheets and Schema Mutable Property Sheets to PAS

also a property schema type registry which is extensible.

"""
from Products.CMFCore.interfaces import ISiteRoot
from Products.PlonePAS.interfaces.propertysheets import IMutablePropertySheet
from Products.PluggableAuthService.UserPropertySheet import UserPropertySheet
from zope.component import getUtility
from zope.interface import implementer


class PropertyValueError(ValueError):
    pass


class PropertySchemaTypeMap:
    def __init__(self):
        self.tmap = {}
        self.tmap_order = []

    def addType(self, type_name, identifier, order=None):
        self.tmap[type_name] = identifier
        if order is not None and isinstance(order, int):
            self.tmap_order.insert(order, type_name)
        else:
            self.tmap_order.append(type_name)

    def getTypeFor(self, value):
        ptypes = [(ptype, self.tmap[ptype]) for ptype in self.tmap_order]
        for ptype, inspector in ptypes:
            if inspector(value):
                return ptype
        raise TypeError("Invalid property type: %s" % type(value))

    def validate(self, property_type, value):
        inspector = self.tmap[property_type]
        return inspector(value)


PropertySchema = PropertySchemaTypeMap()
PropertySchema.addType("string", lambda x: x is None or isinstance(x, str))
PropertySchema.addType("text", lambda x: x is None or isinstance(x, str))
PropertySchema.addType("boolean", lambda x: 1)  # anything can be boolean
PropertySchema.addType("int", lambda x: x is None or isinstance(x, int))
PropertySchema.addType("long", lambda x: x is None or isinstance(x, int))  # theres is no long in Python 3
PropertySchema.addType("float", lambda x: x is None or isinstance(x, float))
PropertySchema.addType("lines", lambda x: x is None or isinstance(x, (tuple, list)))
PropertySchema.addType("selection", lambda x: x is None or isinstance(x, str))
PropertySchema.addType(
    "multiple selection", lambda x: x is None or isinstance(x, (tuple, list))
)
PropertySchema.addType("date", lambda x: 1)
validateValue = PropertySchema.validate


@implementer(IMutablePropertySheet)
class MutablePropertySheet(UserPropertySheet):
    def validateProperty(self, id, value):
        if id not in self._properties:
            raise PropertyValueError("No such property found on this schema")

        proptype = self.getPropertyType(id)
        if not validateValue(proptype, value):
            raise PropertyValueError(
                "Invalid value (%s) for property '%s' of type %s"
                % (value, id, proptype)
            )

    def setProperty(self, user, id, value):
        self.validateProperty(id, value)

        self._properties[id] = value
        self._properties = self._properties

        # cascade to plugin
        provider = self._getPropertyProviderForUser(user)
        provider.setPropertiesForUser(user, self)

    def setProperties(self, user, mapping):
        prop_keys = self._properties.keys()
        prop_update = mapping.copy()

        for key, value in tuple(prop_update.items()):
            if key not in prop_keys:
                prop_update.pop(key)
                continue
            self.validateProperty(key, value)

        self._properties.update(prop_update)

        # cascade to plugin
        provider = self._getPropertyProviderForUser(user)
        provider.setPropertiesForUser(user, self)

    def _getPropertyProviderForUser(self, user):
        # XXX This assumes that the acl_users that we want is in the portal
        # root. This may not always be the case.
        portal = getUtility(ISiteRoot)
        return portal.acl_users._getOb(self._id)


class SchemaMutablePropertySheet(MutablePropertySheet):
    pass



"""
A Local Roles Plugin Implementation that respects Black Listing markers.

ie. containers/objects which denote that they do not wish to acquire local
roles from their containment structure.

"""
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.special_dtml import DTMLFile
from Products.PlonePAS.interfaces.plugins import ILocalRolesPlugin
from Products.PluggableAuthService.plugins.LocalRolePlugin import LocalRolePlugin
from zope.interface import implementer


def manage_addLocalRolesManager(dispatcher, id, title=None, RESPONSE=None):
    """
    add a local roles manager
    """
    lrm = LocalRolesManager(id, title)
    dispatcher._setObject(lrm.getId(), lrm)

    if RESPONSE is not None:
        RESPONSE.redirect("manage_workspace")


manage_addLocalRolesManagerForm = DTMLFile("../zmi/LocalRolesManagerForm", globals())


@implementer(ILocalRolesPlugin)
class LocalRolesManager(LocalRolePlugin):
    """Class incorporating local role storage with
    PlonePAS-specific local role permission checking.
    """

    meta_type = "Local Roles Manager"
    security = ClassSecurityInfo()

    def __init__(self, id, title=None):
        self._id = self.id = id
        self.title = title

    # security.declarePrivate( 'getRolesInContext' )
    def getRolesInContext(self, user, object):
        user_id = user.getId()
        group_ids = user.getGroups()

        principal_ids = list(group_ids)
        principal_ids.insert(0, user_id)

        local = {}
        object = aq_inner(object)

        while 1:
            local_roles = getattr(object, "__ac_local_roles__", None)

            if local_roles and callable(local_roles):
                local_roles = local_roles()

            if local_roles:
                dict = local_roles

                for principal_id in principal_ids:
                    for role in dict.get(principal_id, []):
                        local[role] = 1

            inner = aq_inner(object)
            parent = aq_parent(inner)

            if getattr(object, "__ac_local_roles_block__", None):
                break

            if parent is not None:
                object = parent
                continue

            new = getattr(object, "__self__", None)

            if new is not None:
                object = aq_inner(new)
                continue

            break

        return local.keys()

    # security.declarePrivate('checkLocalRolesAllowed')
    def checkLocalRolesAllowed(self, user, object, object_roles):
        # Still have not found a match, so check local roles. We do
        # this manually rather than call getRolesInContext so that
        # we can incur only the overhead required to find a match.
        inner_obj = aq_inner(object)
        user_id = user.getId()
        group_ids = user.getGroups()

        principal_ids = list(group_ids)
        principal_ids.insert(0, user_id)

        while 1:

            local_roles = getattr(inner_obj, "__ac_local_roles__", None)

            if local_roles and callable(local_roles):
                local_roles = local_roles()

            if local_roles:
                dict = local_roles

                for principal_id in principal_ids:
                    local_roles = dict.get(principal_id, [])

                    # local_roles is empty most of the time, where as
                    # object_roles is usually not.
                    if not local_roles:
                        continue

                    for role in object_roles:
                        if role in local_roles:
                            if user._check_context(object):
                                return 1
                            return 0

            inner = aq_inner(inner_obj)
            parent = aq_parent(inner)

            if getattr(inner_obj, "__ac_local_roles_block__", None):
                break

            if parent is not None:
                inner_obj = parent
                continue

            new = getattr(inner_obj, "__self__", None)

            if new is not None:
                inner_obj = aq_inner(new)
                continue

            break

        return None

    def getAllLocalRolesInContext(self, context):
        roles = {}
        object = aq_inner(context)

        while True:

            local_roles = getattr(object, "__ac_local_roles__", None)

            if local_roles and callable(local_roles):
                local_roles = local_roles()

            if local_roles:

                dict = local_roles

                for principal, localroles in dict.items():
                    if principal not in roles:
                        roles[principal] = set()

                    roles[principal].update(localroles)

            inner = aq_inner(object)
            parent = aq_parent(inner)

            if getattr(object, "__ac_local_roles_block__", None):
                break

            if parent is not None:
                object = parent
                continue

            new = getattr(object, "__self__", None)

            if new is not None:
                object = aq_inner(new)
                continue

            break

        return roles


InitializeClass(LocalRolesManager)



from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import TEST_USER_ID
from plone.testing import zope as zope_testing

import Products.PlonePAS


class ProductsPlonepasLayer(PloneSandboxLayer):

    defaultBases = (PLONE_APP_CONTENTTYPES_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base
        # layer.
        self.loadZCML(package=Products.PlonePAS)
        zope_testing.installProduct(app, "Products.PlonePAS")

    def setUpPloneSite(self, portal):
        applyProfile(portal, "Products.PlonePAS:PlonePAS")
        # setRoles(portal, TEST_USER_ID, ['Manager'])
        from Products.CMFPlone.utils import _createObjectByType

        _createObjectByType("Folder", portal, id="Members")
        mtool = portal.portal_membership
        if not mtool.getMemberareaCreationFlag():
            mtool.setMemberareaCreationFlag()
        mtool.createMemberArea(TEST_USER_ID)
        if mtool.getMemberareaCreationFlag():
            mtool.setMemberareaCreationFlag()

        _createObjectByType("Folder", portal, id="folder")


PRODUCTS_PLONEPAS_FIXTURE = ProductsPlonepasLayer()


PRODUCTS_PLONEPAS_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PRODUCTS_PLONEPAS_FIXTURE,),
    name="ProductsPlonepasLayer:IntegrationTesting",
)


PRODUCTS_PLONEPAS_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PRODUCTS_PLONEPAS_FIXTURE,),
    name="ProductsPlonepasLayer:FunctionalTesting",
)



"""
Password Validation plugin (IValidationPlugin)
Mutable Property Provider
"""
from AccessControl import ClassSecurityInfo
from AccessControl.class_init import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PluggableAuthService.interfaces.plugins import IValidationPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from zope.i18nmessageid import MessageFactory
from zope.interface import implementer


_ = MessageFactory("plone")

manage_addPasswordPolicyForm = PageTemplateFile("../zmi/PasswordPolicyForm", globals())


def manage_addPasswordPolicyPlugin(
    self, id, title="", RESPONSE=None, schema=None, **kw
):
    """
    Create an instance of a password validation plugin.
    """
    o = PasswordPolicyPlugin(id, title)
    self._setObject(o.getId(), o)

    if RESPONSE is not None:
        return RESPONSE.redirect(
            "%s/manage_workspace?"
            "manage_tabs_message=DefaultPasswordPlugin+plugin+added"
            % self.absolute_url()
        )


@implementer(IValidationPlugin)
class PasswordPolicyPlugin(BasePlugin):
    """Simple Password Policy to enforce a minimum password length."""

    meta_type = "Default Plone Password Policy"

    security = ClassSecurityInfo()

    min_chars = 8

    def __init__(self, id, title=""):
        """Create a default plone password policy"""
        self.id = id
        self.title = title

    @security.private
    def validateUserInfo(self, user, set_id, set_info):
        """See IValidationPlugin. Used to validate password property"""

        if not set_info:
            return []
        password = set_info.get("password", None)
        if password is None:
            return []
        elif password == "":
            return [
                {
                    "id": "password",
                    "error": _(
                        "Minimum ${min_chars} characters.",
                        mapping={"min_chars": str(self.min_chars)},
                    ),
                }
            ]
        elif len(password) < self.min_chars:
            return [
                {
                    "id": "password",
                    "error": _(
                        "Your password must contain at least ${min_chars} characters.",
                        mapping={"min_chars": str(self.min_chars)},
                    ),
                }
            ]
        else:
            return []


InitializeClass(PasswordPolicyPlugin)



from Products.PluggableAuthService.interfaces.propertysheets import IPropertySheet


class IMutablePropertySheet(IPropertySheet):
    def canWriteProperty(object, id):
        """Check if a property can be modified."""

    def setProperty(object, id, value):
        """ """

    def setProperties(object, mapping):
        """ """


class ISchemaMutablePropertySheet(IMutablePropertySheet):
    pass



from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING
from Products.PluggableAuthService.interfaces.authservice import IPluggableAuthService
from Products.PluggableAuthService.interfaces.events import IPrincipalDeletedEvent
from Products.PluggableAuthService.interfaces.plugins import IRolesPlugin
from Products.PluggableAuthService.PluggableAuthService import (
    _SWALLOWABLE_PLUGIN_EXCEPTIONS,
)
from zope.component import adapter
from zope.component import getGlobalSiteManager

import unittest


class BasicOpsTestCase(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.acl_users = self.portal.acl_users

    def compareRoles(self, target, user, roles):
        """
        compareRoles(self, target, user, roles) => do not raise if
        user has exactly the specified roles.

        If target is None, test user roles (no local roles)
        """
        u = self.acl_users.getUser(user)
        if not u:
            raise RuntimeError("compareRoles: Invalid user: '%s'" % user)
        non_roles = ("Authenticated", "Anonymous", "")
        if target is None:
            user_roles = list(u.getRoles())
        else:
            user_roles = list(u.getRolesInContext(target))
        actual_roles = list(filter(lambda x: x not in non_roles, user_roles))
        wished_roles = list(roles)
        if sorted(actual_roles) == sorted(wished_roles):
            return 1
        raise RuntimeError(
            "User %s: Whished roles: %s BUT current "
            "roles: %s" % (user, wished_roles, actual_roles)
        )

    def createUser(
        self, login="created_user", password="secret", roles=[], groups=[], domains=()
    ):
        self.acl_users.userFolderAddUser(
            login,
            password,
            roles=roles,
            groups=groups,
            domains=domains,
        )

    def test_installed(self):
        self.assertTrue(IPluggableAuthService.providedBy(self.acl_users))

    def test_add(self):
        self.createUser()
        self.assertTrue(self.acl_users.getUser("created_user"))

    def test_edit(self):
        # this will fail unless the PAS role plugin is told it manages
        # the new role.
        self.createUser()
        self.compareRoles(None, "created_user", [])
        self.acl_users.userFolderEditUser(
            "created_user",  # name
            "secret2",  # password
            roles=["Member"],
            groups=["g1"],
            domains=(),
        )
        self.compareRoles(None, "created_user", ["Member"])

    def test_edit_userDefinedRole(self):
        roleplugins = self.acl_users.plugins.listPlugins(IRolesPlugin)
        for id, plugin in roleplugins:
            try:
                plugin.addRole("r1")
            except _SWALLOWABLE_PLUGIN_EXCEPTIONS:
                pass
            else:
                break

        self.createUser()
        self.compareRoles(None, "created_user", [])
        self.acl_users.userFolderEditUser(
            "created_user",  # name
            "secret2",  # password
            roles=["r1"],
            groups=["g1"],
            domains=(),
        )
        self.compareRoles(None, "created_user", ["r1"])

    def test_del(self):
        self.createUser()
        self.assertTrue(self.acl_users.getUser("created_user"))
        self.acl_users.userFolderDelUsers(["created_user"])
        self.assertFalse(self.acl_users.getUser("created_user"))

    def test_principal_del_event(self):
        eventsFired = []

        @adapter(IPrincipalDeletedEvent)
        def gotDeletion(event):
            eventsFired.append(event)

        gsm = getGlobalSiteManager()
        gsm.registerHandler(gotDeletion)
        self.createUser()
        self.acl_users.userFolderDelUsers(["created_user"])
        self.assertEqual(len(eventsFired), 1)
        self.assertEqual(eventsFired[0].principal, "created_user")
        gsm.unregisterHandler(gotDeletion)

    def test_search(self):
        self.createUser("created_user1")
        self.createUser("created_user2")
        mt = self.portal.portal_membership
        retlist = mt.searchForMembers(REQUEST=None, login="created_user1")
        usernames = [user.getUserName() for user in retlist]
        self.assertEqual(len(usernames), 1)
        self.assertTrue(
            "created_user1" in usernames, "'created_user1' not in %s" % usernames
        )

    def test_setpw(self):
        # there is more than one place where one can set the password.
        # insane. anyway we have to check the patch in pas.py userSetPassword
        # here its checked in the general setup using ZODBUserManager.
        self.createUser()
        uf = self.acl_users
        new_secret = "new_secret"
        uf.userSetPassword("created_user", new_secret)

        # possible to authenticate with new password?
        from Products.PluggableAuthService.interfaces.plugins import (
            IAuthenticationPlugin,
        )

        authenticators = uf.plugins.listPlugins(IAuthenticationPlugin)
        credentials = {"login": "created_user", "password": new_secret}
        result = None
        for aid, authenticator in authenticators:
            result = authenticator.authenticateCredentials(credentials)
            if result is not None:
                break
        self.assertTrue(result)

    def test_setProperties(self):
        self.createUser()
        user = self.acl_users.getUser("created_user")
        user.setProperties(fullname="Test User", email="test@example.org")
        self.assertEqual(user.getProperty("fullname"), "Test User")
        self.assertEqual(user.getProperty("email"), "test@example.org")

        user.setProperties(
            properties={"fullname": "Test User2", "email": "test2@example.org"}
        )
        self.assertEqual(user.getProperty("fullname"), "Test User2")
        self.assertEqual(user.getProperty("email"), "test2@example.org")



# Define certain interfaces that a plugin must meet if it is to allow
# certain operations to be done by the Plone UI.
from Products.PluggableAuthService.interfaces.plugins import Interface


class IDeleteCapability(Interface):
    """Interface for plugin to say if it allows for deletion of a user.
    Expected to be found on IUserManagement plugins.
    For Plone UI.
    """

    def allowDeletePrincipal(id):
        """True iff this plugin can delete a certain user/group."""


class IPasswordSetCapability(Interface):
    """Interface for plugin to say if it allows for setting the password of a
    user.
    Expected to be found on IUserManagement plugins.
    For Plone UI.
    """

    def allowPasswordSet(id):
        """True iff this plugin can set the password of a certain user."""


class IGroupCapability(Interface):
    """Interface for plugin to say if it allows for deletion of a user.
    For Plone UI.
    """

    def allowGroupAdd(principal_id, group_id):
        """True iff this plugin will allow adding a certain principal to a
        certain group."""

    def allowGroupRemove(principal_id, group_id):
        """True iff this plugin will allow removing a certain principal from a
        certain group."""


class IAssignRoleCapability(Interface):
    """Interface for plugin to say if it allows for deletion of a user.
    For Plone UI.
    """

    def allowRoleAssign(id, role):
        """True iff this plugin will allow assigning a certain principal a
        certain role."""


class IManageCapabilities(Interface):
    """Interface for MemberData/GroupData to provide information as to whether
    or not the member can be deleted, reset password, modify a property.
    """

    def canDelete():
        """True iff user can be removed from the Plone UI."""

    def canPasswordSet():
        """True iff user can change password."""

    def passwordInClear():
        """True iff password can be retrieved in the clear (not hashed.)"""

    def canWriteProperty(prop_name):
        """True iff the member/group property named in 'prop_name'
        can be changed.
        """

    def canAddToGroup(group_id):
        """True iff member can be added to group."""

    def canRemoveFromGroup(group_id):
        """True iff member can be removed from group."""

    def canAssignRole(role):
        """True iff member can be assigned role."""



""" Class: CookieCrumblingPlugin

Acts as auth plugin, but injects cookie form credentials as HTTPBasicAuth.
This allows form logins to fall through to parent user folders.

"""
from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import aq_base
from App.special_dtml import DTMLFile
from OFS.Folder import Folder
from Products.CMFCore.CookieCrumbler import manage_addCC
from Products.PluggableAuthService.interfaces.plugins import IExtractionPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from zope.interface import implementer

import logging


logger = logging.getLogger("PlonePAS")

CC_ID = "cookie_auth"


def manage_addCookieCrumblingPlugin(self, id, title="", RESPONSE=None, **kw):
    """
    Create an instance of a cookie crumbling plugin.
    """
    self = self.this()

    o = CookieCrumblingPlugin(id, title, **kw)
    self._setObject(o.getId(), o)
    o = getattr(aq_base(self), id)

    manage_addCC(o, CC_ID)

    if RESPONSE is not None:
        RESPONSE.redirect("manage_workspace")


manage_addCookieCrumblingPluginForm = DTMLFile(
    "../zmi/CookieCrumblingPluginForm", globals()
)


@implementer(IExtractionPlugin)
class CookieCrumblingPlugin(Folder, BasePlugin):
    """Multi-plugin for injecting HTTP Basic Authentication
    credentials from form credentials.
    """

    meta_type = "Cookie Crumbling Plugin"

    security = ClassSecurityInfo()

    def __init__(self, id, title=None):
        self._setId(id)
        self.title = title

    def _getCC(self):
        return getattr(aq_base(self), CC_ID, None)

    @security.private
    def extractCredentials(self, request):
        """Extract basic auth credentials from 'request'."""

        try:
            self._getCC().modifyRequest(request, request.RESPONSE)

        except Exception as e:
            logger.error("PlonePAS error: %s", e, exc_info=1)

        return {}


InitializeClass(CookieCrumblingPlugin)



"""
Test set up specific to Plone through thea GenericSetup profile installation.
"""

from plone.app import testing as pa_testing
from plone.testing import zope
from Products.PlonePAS import testing
from Products.PluggableAuthService.interfaces import plugins as plugins_ifaces
from Products.PluggableAuthService.plugins import CookieAuthHelper
from Products.PluggableAuthService.plugins import HTTPBasicAuthHelper
from zope.component import hooks

import transaction
import unittest
import urllib.parse


class PortalSetupTest(unittest.TestCase):
    """
    Test set up specific to Plone through thea GenericSetup profile installation.
    """

    layer = testing.PRODUCTS_PLONEPAS_FUNCTIONAL_TESTING

    def setUp(self):
        """
        Set up convenience references to test fixture from the layer.
        """
        self.app = self.layer["app"]
        self.root_acl_users = self.app.acl_users
        self.portal = self.layer["portal"]

    def test_zope_root_default_challenge(self):
        """
        The Zope root `/acl_users` default challenge plugin works.
        """
        # Check the Zope root PAS plugin configuration
        self.assertIn(
            "credentials_basic_auth",
            self.root_acl_users.objectIds(),
            "Basic auth plugin missing from Zope root `/acl_users`",
        )
        basic_plugin = self.root_acl_users.credentials_basic_auth
        self.assertIsInstance(
            basic_plugin,
            HTTPBasicAuthHelper.HTTPBasicAuthHelper,
            "Wrong Zope root `/acl_users` basic auth plugin type",
        )
        challenge_plugins = self.root_acl_users.plugins.listPlugins(
            plugins_ifaces.IChallengePlugin,
        )
        _, default_challenge_plugin = challenge_plugins[0]
        self.assertEqual(
            "/".join(default_challenge_plugin.getPhysicalPath()),
            "/".join(basic_plugin.getPhysicalPath()),
            "Wrong Zope root `/acl_users` default challenge plugin",
        )

        # Check the challenge response in the actual browser
        browser = zope.Browser(self.app)
        browser.raiseHttpErrors = False
        browser.open(self.app.absolute_url() + "/manage_main")
        self.assertEqual(
            browser.headers["Status"].lower(),
            "401 unauthorized",
            "Wrong Zope root `/acl_users` default challenge response status",
        )

    def test_zope_root_cookie_login(self):
        """
        The Zope root `/acl_users` cookie login works.
        """
        # Install the GenericSetup profile that performs the actual switch
        pa_testing.applyProfile(self.portal, "Products.PlonePAS:root-cookie")
        transaction.commit()

        # Make the cookie plugin the default auth challenge
        self.assertIn(
            "credentials_cookie_auth",
            self.root_acl_users.objectIds(),
            "Cookie auth plugin missing from Zope root `/acl_users`",
        )
        cookie_plugin = self.root_acl_users.credentials_cookie_auth
        self.assertIs(
            type(cookie_plugin.aq_base),
            CookieAuthHelper.CookieAuthHelper,
            "Wrong Zope root `/acl_users` cookie auth plugin type",
        )
        challenge_plugins = self.root_acl_users.plugins.listPlugins(
            plugins_ifaces.IChallengePlugin,
        )
        _, default_challenge_plugin = challenge_plugins[0]
        self.assertEqual(
            "/".join(default_challenge_plugin.getPhysicalPath()),
            "/".join(cookie_plugin.getPhysicalPath()),
            "Wrong Zope root `/acl_users` default challenge plugin",
        )

        # Check the challenge response in the actual browser
        browser = zope.Browser(self.app)
        browser.open(self.app.absolute_url() + "/manage_main")
        self.assertEqual(
            browser.headers["Status"].lower(),
            "200 ok",
            "Wrong Zope root `/acl_users` cookie challenge response status",
        )
        login_form_url = urllib.parse.urlsplit(browser.url)
        self.assertEqual(
            login_form_url._replace(query="").geturl(),
            cookie_plugin.absolute_url() + "/login_form",
            "Wrong Zope root `/acl_users` cookie challenge redirect",
        )

        # Workaround the fact that the `zope.component` site is still the Plone portal
        # when the test browser handles requests.
        hooks.setSite(None)
        zope.login(self.root_acl_users, pa_testing.SITE_OWNER_NAME)
        self.app.REQUEST.form["__ac_name"] = pa_testing.SITE_OWNER_NAME
        self.app.REQUEST.form["__ac_password"] = pa_testing.TEST_USER_PASSWORD
        cookie_plugin.login()

        # Submit the login form in the browser
        login_form = browser.getForm()
        login_form.getControl(name="__ac_name").value = pa_testing.SITE_OWNER_NAME
        login_form.getControl(
            name="__ac_password"
        ).value = pa_testing.TEST_USER_PASSWORD
        login_form.controls[-1].click()
        self.assertEqual(
            browser.headers["Status"].lower(),
            "200 ok",
            "Wrong Zope root `/acl_users` cookie login response status",
        )
        self.assertEqual(
            browser.url,
            self.app.absolute_url() + "/manage_main",
            "Wrong Zope root `/acl_users` cookie login redirect",
        )



from Products.PlonePAS.testing import PRODUCTS_PLONEPAS_INTEGRATION_TESTING

import unittest


class TestPASSearchView(unittest.TestCase):

    layer = PRODUCTS_PLONEPAS_INTEGRATION_TESTING

    def test_sort(self):
        self.portal = self.layer["portal"]
        pas_search = self.portal.restrictedTraverse("@@pas_search")
        values = [
            {"title": "Sociologie"},
            {"title": "Économie"},
            {"title": "anthropologie"},
        ]
        sorted_values = pas_search.sort(values, "title")
        # do not modify original
        self.assertEqual(
            values,
            [
                {"title": "Sociologie"},
                {"title": "Économie"},
                {"title": "anthropologie"},
            ],
        )
        # sorted here
        self.assertEqual(
            sorted_values,
            [
                {"title": "anthropologie"},
                {"title": "Économie"},
                {"title": "Sociologie"},
            ],
        )



from Acquisition import aq_inner
from plone.memoize.instance import memoize
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.PlonePAS.interfaces.browser import IPASInfoView
from Products.PluggableAuthService.interfaces.plugins import IExtractionPlugin
from Products.PluggableAuthService.interfaces.plugins import (  # noqa: E501
    ILoginPasswordExtractionPlugin,
)
from zope.interface import implementer


@implementer(IPASInfoView)
class PASInfoView(BrowserView):
    def checkExtractorForInterface(self, interface):
        acl = getToolByName(aq_inner(self.context), "acl_users")
        plugins = acl.plugins.listPlugins(IExtractionPlugin)

        for plugin in plugins:
            if interface.providedBy(plugin[1]):
                return True

        return False

    @memoize
    def hasLoginPasswordExtractor(self):
        return self.checkExtractorForInterface(ILoginPasswordExtractionPlugin)

    @memoize
    def hasOpenIDExtractor(self):
        try:
            from plone.openid.interfaces import IOpenIdExtractionPlugin
        except ImportError:
            return False

        return self.checkExtractorForInterface(IOpenIdExtractionPlugin)

    def hasOpenIDdExtractor(self):
        # BBB Keeping method name with typo for backwards compatibility.
        return self.hasOpenIDExtractor()



""" Class: ExtendedCookieAuthHelper

Simply extends the standard CookieAuthHelper provided via regular
PluggableAuthService but overrides the updateCookie mechanism to
provide similar functionality as CookieCrumbler does... by giving
the portal the ability to provide a setAuthCookie method.
"""
from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from Acquisition import aq_base
from Acquisition import aq_parent
from App.special_dtml import DTMLFile
from DateTime import DateTime
from plone.registry.interfaces import IRegistry
from Products.PluggableAuthService.interfaces.authservice import IPluggableAuthService
from Products.PluggableAuthService.interfaces.plugins import IChallengePlugin
from Products.PluggableAuthService.interfaces.plugins import ICredentialsResetPlugin
from Products.PluggableAuthService.interfaces.plugins import ICredentialsUpdatePlugin
from Products.PluggableAuthService.interfaces.plugins import (
    ILoginPasswordHostExtractionPlugin,
)
from Products.PluggableAuthService.plugins.CookieAuthHelper import (
    CookieAuthHelper as BasePlugin,
)
from urllib.parse import quote
from zope.component import getUtility
from zope.interface import implementer


def manage_addExtendedCookieAuthHelper(self, id, title="", RESPONSE=None, **kw):
    """Create an instance of a extended cookie auth helper."""

    self = self.this()

    o = ExtendedCookieAuthHelper(id, title, **kw)
    self._setObject(o.getId(), o)
    o = getattr(aq_base(self), id)

    if RESPONSE is not None:
        RESPONSE.redirect("manage_workspace")


manage_addExtendedCookieAuthHelperForm = DTMLFile(
    "../zmi/ExtendedCookieAuthHelperForm", globals()
)


@implementer(
    ILoginPasswordHostExtractionPlugin,
    IChallengePlugin,
    ICredentialsUpdatePlugin,
    ICredentialsResetPlugin,
)
class ExtendedCookieAuthHelper(BasePlugin):
    """Multi-plugin which adds ability to override the updating of cookie via
    a setAuthCookie method/script.
    """

    meta_type = "Extended Cookie Auth Helper"
    security = ClassSecurityInfo()

    @security.private
    def updateCredentials(self, request, response, login, new_password):
        """Override standard updateCredentials method"""
        cookie_val = self.get_cookie_value(login, new_password)
        kw = {}
        registry = getUtility(IRegistry)
        length = registry.get("plone.auth_cookie_length", "0")
        try:
            length = int(length)
        except ValueError:
            length = 0
        if length:
            kw.update(expires=(DateTime() + length).toZone("GMT").rfc822())
        response.setCookie(self.cookie_name, quote(cookie_val), path="/", **kw)

    @security.public
    def login(self):
        """Set a cookie and redirect to the url that we tried to
        authenticate against originally.

        Override standard login method to avoid calling
        'return response.redirect(came_from)' as there is additional
        processing to ignore known bad come_from templates at
        login_next.cpy script.
        """
        request = self.REQUEST
        response = request["RESPONSE"]

        password = request.get("__ac_password", "")

        user = getSecurityManager().getUser()
        login = user.getUserName()
        user_pas = aq_parent(user)

        if IPluggableAuthService.providedBy(user_pas):
            # Delegate to the users own PAS if possible
            user_pas.updateCredentials(request, response, login, password)
        else:
            # User does not originate from a PAS user folder, so lets try
            # to do our own thing.
            # XXX Perhaps we should do nothing here; test with pure User
            # Folder!
            pas_instance = self._getPAS()
            if pas_instance is not None:
                pas_instance.updateCredentials(request, response, login, password)


InitializeClass(ExtendedCookieAuthHelper)



from setuptools import find_packages
from setuptools import setup

import sys


version = "7.0.0b2"


longdescription = open("README.rst").read()
longdescription += "\n"
longdescription += open("CHANGES.rst").read()

install_requires = [
    "plone.base",
    "plone.i18n",
    "plone.memoize",
    "plone.protect>=2.0.3",
    "plone.session",
    "Products.GenericSetup",
    "Products.PluggableAuthService>=2.0b2.dev0",
    "setuptools",
]

setup(
    name="Products.PlonePAS",
    version=version,
    description="PlonePAS modifies the PluggableAuthService for use by Plone.",
    long_description=longdescription,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: Zope Public License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="Zope CMF Plone PAS authentication",
    author="Kapil Thangavelu, Wichert Akkerman",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://github.com/plone/Products.PlonePAS",
    license="ZPL",
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["Products"],
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require=dict(
        test=[
            "plone.app.testing",
            "plone.testing",
            "plone.app.robotframework",
        ],
    ),
)



from plone.i18n.normalizer.interfaces import IIDNormalizer
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.PlonePAS.interfaces.browser import IPASSearchView
from zope.component import queryUtility
from zope.interface import implementer


@implementer(IPASSearchView)
class PASSearchView(BrowserView):
    @staticmethod
    def extractCriteriaFromRequest(request):
        criteria = request.form.copy()

        for key in ["form.submitted", "submit", "b_start", "b_size"]:
            if key in criteria:
                del criteria[key]

        for (key, value) in criteria.items():
            if not value:
                del criteria[key]

        return criteria

    @staticmethod
    def merge(results, key):
        output = {}
        for entry in results:
            id = entry[key]
            if id not in output:
                output[id] = entry.copy()
            else:
                buf = entry.copy()
                buf.update(output[id])
                output[id] = buf

        return output.values()

    def sort(self, results, sort_key):
        idnormalizer = queryUtility(IIDNormalizer)

        def key_func(a):
            return idnormalizer.normalize(a.get(sort_key, a))

        return sorted(results, key=key_func)

    def searchUsers(self, sort_by=None, **criteria):
        self.pas = getToolByName(self.context, "acl_users")
        results = self.merge(self.pas.searchUsers(**criteria), "userid")
        if sort_by is not None:
            results = self.sort(results, sort_by)
        return results

    def searchUsersByRequest(self, request, sort_by=None):
        criteria = self.extractCriteriaFromRequest(request)
        return self.searchUsers(sort_by=sort_by, **criteria)

    def searchGroups(self, sort_by=None, **criteria):
        self.pas = getToolByName(self.context, "acl_users")
        results = self.merge(self.pas.searchGroups(**criteria), "groupid")
        if sort_by is not None:
            results = self.sort(results, sort_by)
        return results

    def searchGroupsByRequest(self, request):
        criteria = self.extractCriteriaFromRequest(request)
        return self.searchGroups(**criteria)

    def getPhysicalPath(self):
        # We call various PAS methods which can be ZCached. The ZCache
        # infrastructure relies on getPhysicalPath on the context being
        # available, which this view does not have, it not being a
        # persistent object. So we fake things and return the physical path
        # for our context.
        return self.context.getPhysicalPath()



from zope.interface import Interface


class IPASInfoView(Interface):
    def hasLoginPasswordExtractor():
        """Check if a login & password extraction plugin is active.

        Check if there is a plugin with an enabled
        ILoginPasswordExtractionPlugin interface. This can be used to
        conditionally show username & password logins.
        """

    def hasOpenIDExtractor():
        """Check if an OpenID extraction plugin is active."""

    def hasOpenIDdExtractor():
        """Check if an OpenID extraction plugin is active.

        BBB Keeping method name with typo for backwards compatibility.
        """


class IPASMemberView(Interface):
    def info(userid=None):
        """Return 'harmless' member info of any user, such as full name,
        location, etc.
        """


class IPASSearchView(Interface):
    def searchUsers(sort_by=None, any_field=None, **criteria):
        """Search for users matching a set of criteria.

        The criteria are a dictionary mapping user properties to values and
        have the semantics declared by IPluggableAuthService.searchUsers().
        Duplicate results returned by PAS are filtered so only the first
        result remains in the result set. The results can be sorted on
        sort_bys (case insensitive).

        In addition, a single search string can be sought in multiple
        fields at a time by passing any_field='your-string'. The fields
        that will be searched are not explicitly enumerated but will be the
        user-facing ones a user would expect to search, such as login name
        and full name. (Perhaps any_field should/could be moved into PAS
        proper. This could be done without breaking code that depends on
        this interface.)

        If you specify both any_field and other criteria that include a
        field that any_field would typically search (such as login name),
        the other criteria will be enforced at the expense of any_field.
        """

    def searchUsersByRequest(request, sort_by=None):
        """Search for users matching a set of criteria found in a request.

        This method will look remove any obvious values from the request
        which are not search criteria. It will also remove any fields
        which have an empty string value.
        Duplicate results returned by PAS are filtered so only the first
        result remains in the result set. The results can be sorted on
        sort_by (case insensitive).
        """

    def searchGroups(**criteria):
        """Search for groups matching a set of criteria.

        The criteria are a dictionary mapping group properties
        to values and have the semantics declared by
        IPluggableAuthService.searchGroups().
        """

    def searchGroupsByRequest(request):
        """Search for groups matching a set of criteria found in a request.

        This method will look remove any obvious values from the request
        which are not search criteria. It will also remove any fields
        which have an empty string value.
        """

    def merge(results, key):
        """merge two search results based on key as the unique criterion"""

    def sort(results, key):
        """sort results on a key"""
